# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from math import exp, sqrt, tanh
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common.decorators import Registry


registry: Registry[tp.Callable[[np.ndarray], float]] = Registry()


class BonnansFunction:
    def __init__(self, index: int, M: int = 100, N: int = 100) -> None:
        self.N = N
        self.M = M
        self.A = np.random.RandomState(index).rand(M, N)
        self.y = np.random.RandomState(index + 1).rand(M) * N / 2
        # print(f"y={self.y} {sum(self.y)} {sum((self.y/self.N)**2)}")

    def __call__(self, x: tp.ArrayLike) -> float:
        assert len(x) == self.N
        return np.sum((np.matmul(self.A, x) / self.N - self.y / self.N) ** 2)


class DiscreteFunction:
    def __init__(self, name: str, arity: int = 2) -> None:
        """Returns a classical discrete function for test, in the domain {0,1,...,arity-1}^d.
        The name can be onemax, leadingones, or jump.

        onemax(x) is the most classical case of discrete functions, adapted to minimization.
        It is originally designed for lists of bits. It just counts the number of 1,
        and returns len(x) - number of ones. However, the present function perturbates the location of the
        optimum, so that tests can not be easily biased by a wrong initialization. So the optimum,
        instead of being located at (1,1,...,1), is located at (0,1,2,...,arity-1,0,1,2,...).

        leadingones is the second most classical discrete function, adapted for minimization.
        Before perturbation of the location of the optimum as above,
        it returns len(x) - number of initial 1. I.e.
        leadingones([0 1 1 1]) = 4,
        leadingones([1 1 1 1]) = 0,
        leadingones([1 0 0 0]) = 3.
        The present Leadingones function uses a perturbation as documented above for OneMax: we count the number
        of initial correct values, a correct values being 0 for variable 1, 1 for variable 2, 2 for variable 3, and
        so on.

        There exists variants of jump functions: the principle of a jump function is that local descent does not succeed.
        Jumps are necessary. We are here in minimization, hence a formulation slightly different from most discrete optimization
        papers, which usually assume maximization. We use the same perturbation as detailed above for leadingones and onemax,
        i.e. the optimum is located at (0,1,2,...,arity-1,0,1,2,...).
        """
        self._arity = arity
        self._func = dict(onemax=self.onemax, leadingones=self.leadingones, jump=self.jump)[name]

    def __call__(self, x: tp.ArrayLike) -> float:
        return self._func(x)

    def onemax(self, x: tp.ArrayLike) -> float:
        diff = np.round(x) - (np.arange(len(x)) % self._arity)
        return float(np.sum(diff != 0))

    def leadingones(self, x: tp.ArrayLike) -> float:
        diff = np.round(x) - (np.arange(len(x)) % self._arity)
        nonzeros = np.nonzero(diff)[0]
        return float(len(x) - nonzeros[0] if nonzeros.size else 0)

    def jump(self, x: tp.ArrayLike) -> float:
        n = len(x)
        m = n // 4
        o = n - self.onemax(x)
        if o == n or o <= n - m:
            return n - m - o
        return o  # Deceptive part.


def _styblinksitang(x: np.ndarray, noise: float) -> float:
    """Classical function for testing noisy optimization."""
    x2 = x**2
    val = x2.dot(x2) + np.sum(5 * x - 16 * x2)
    # return a positive value for maximization
    return float(39.16599 * len(x) + 0.5 * val + noise * np.random.normal(size=val.shape))


class DelayedSphere:
    def __call__(self, x: np.ndarray) -> float:
        return float(np.sum(x**2))

    def compute_pseudotime(  # pylint: disable=unused-argument
        self, input_parameter: tp.Any, value: float
    ) -> float:
        x = input_parameter[0][0]
        return float(abs(1.0 / x[0]) / 1000.0) if x[0] != 0.0 else 0.0


registry.register(DelayedSphere())


@registry.register
def sphere(x: np.ndarray) -> float:
    """The most classical continuous optimization testbed.

    If you do not solve that one then you have a bug."""
    assert x.ndim == 1
    return float(x.dot(x))


@registry.register
def sphere1(x: np.ndarray) -> float:
    """Translated sphere function."""
    return sphere(x - 1.0)


@registry.register
def sphere2(x: np.ndarray) -> float:
    """A bit more translated sphere function."""
    return sphere(x - 2.0)


@registry.register
def sphere4(x: np.ndarray) -> float:
    """Even more translated sphere function."""
    return sphere(x - 4.0)


@registry.register
def maxdeceptive(x: np.ndarray) -> float:
    dec = 3 * x**2 - (2 / (3 ** (x - 2) ** 2 + 0.1))
    return float(np.max(dec))


@registry.register
def sumdeceptive(x: np.ndarray) -> float:
    dec = 3 * x**2 - (2 / (3 ** (x - 2) ** 2 + 0.1))
    return float(np.sum(dec))


@registry.register
def altcigar(x: np.ndarray) -> float:
    """Similar to cigar, but variables in inverse order.

    E.g. for pointing out algorithms not invariant to the order of variables."""
    return float(x[-1]) ** 2 + 1000000.0 * sphere(x[:-1])


@registry.register
def discus(x: np.ndarray) -> float:
    """Only one variable is very penalized."""
    return sphere(x[1:]) + 1000000.0 * float(x[0]) ** 2


@registry.register
def cigar(x: np.ndarray) -> float:
    """Classical example of ill conditioned function.

    The other classical example is ellipsoid.
    """
    return float(x[0]) ** 2 + 1000000.0 * sphere(x[1:])


@registry.register
def bentcigar(x: np.ndarray) -> float:
    """Classical example of ill conditioned function, but bent."""
    y = np.asarray(
        [
            x[i] ** (1 + 0.5 * np.sqrt(x[i]) * (i - 1) / (len(x) - 1)) if x[i] > 0.0 else x[i]
            for i in range(len(x))
        ]
    )
    return float(y[0]) ** 2 + 1000000.0 * sphere(y[1:])


@registry.register
def multipeak(x: np.ndarray) -> float:
    """Inspired by M. Gallagher's Gaussian peaks function."""
    v = 10000.0
    for a in range(101):
        x_ = np.asarray([np.cos(a + np.sqrt(i)) for i in range(len(x))])
        v = min(v, a / 101.0 + np.exp(sphere(x - x_)))
    return v


@registry.register
def altellipsoid(y: np.ndarray) -> float:
    """Similar to Ellipsoid, but variables in inverse order.

    E.g. for pointing out algorithms not invariant to the order of variables."""
    return ellipsoid(y[::-1])


def step(s: float) -> float:
    if s == 0.0:
        s = np.nextafter(0.0, 1.0)
    return float(np.exp(int(np.log(s))))


@registry.register
def stepellipsoid(x: np.ndarray) -> float:
    """Classical example of ill conditioned function.

    But we add a 'step', i.e. we set the gradient to zero everywhere.
    Compared to some existing testbeds, we decided to have infinitely many steps.
    """
    dim = x.size
    weights = 10 ** np.linspace(0, 6, dim)
    return float(step(weights.dot(x**2)))


@registry.register
def ellipsoid(x: np.ndarray) -> float:
    """Classical example of ill conditioned function.

    The other classical example is cigar.
    """
    dim = x.size
    weights = 10 ** np.linspace(0, 6, dim)
    return float(weights.dot(x**2))


@registry.register
def rastrigin(x: np.ndarray) -> float:
    """Classical multimodal function."""
    cosi = float(np.sum(np.cos(2 * np.pi * x)))
    return float(10 * (len(x) - cosi) + sphere(x))


@registry.register
def bucherastrigin(x: np.ndarray) -> float:
    """Classical multimodal function. No box-constraint penalization here."""
    s = np.asarray(
        [
            x[i] * (10 if x[i] > 0.0 and i % 2 else 1) * (10 ** ((i - 1) / (2 * (len(x) - 1))))
            for i in range(len(x))
        ]
    )
    cosi = float(np.sum(np.cos(2 * np.pi * s)))
    return float(10 * (len(x) - cosi) + sphere(s))


@registry.register
def doublelinearslope(x: np.ndarray) -> float:
    """We decided to use two linear slopes rather than having a constraint artificially added for
    not having the optimum at infinity."""
    return float(np.abs(np.sum(x)))


@registry.register
def stepdoublelinearslope(x: np.ndarray) -> float:
    return step(np.abs(np.sum(x)))


@registry.register
def hm(x: np.ndarray) -> float:
    """New multimodal function (proposed for Nevergrad)."""
    return float((x**2).dot(1.1 + np.cos(1.0 / x)))


@registry.register
def rosenbrock(x: np.ndarray) -> float:
    x_m_1 = x[:-1] - 1
    x_diff = x[:-1] ** 2 - x[1:]
    return float(100 * x_diff.dot(x_diff) + x_m_1.dot(x_m_1))


@registry.register
def ackley(x: np.ndarray) -> float:
    dim = x.size
    sum_cos = np.sum(np.cos(2 * np.pi * x))
    return -20.0 * exp(-0.2 * sqrt(sphere(x) / dim)) - exp(sum_cos / dim) + 20 + exp(1)


@registry.register
def schwefel_1_2(x: np.ndarray) -> float:
    cx = np.cumsum(x)
    return sphere(cx)


@registry.register
def griewank(x: np.ndarray) -> float:
    """Multimodal function, often used in Bayesian optimization."""
    part1 = sphere(x)
    part2 = np.prod(np.cos(x / np.sqrt(1 + np.arange(len(x)))))
    return 1 + (float(part1) / 4000.0) - float(part2)


@registry.register
def deceptiveillcond(x: np.ndarray) -> float:
    """An extreme ill conditioned functions. Most algorithms fail on this.

    The condition number increases to infinity as we get closer to the optimum."""
    assert len(x) >= 2
    return float(
        max(np.abs(np.arctan(x[1] / x[0])), np.sqrt(x[0] ** 2.0 + x[1] ** 2.0), 1.0 if x[0] > 0 else 0.0)
        if x[0] != 0.0
        else float("inf")
    )


@registry.register
def deceptivepath(x: np.ndarray) -> float:
    """A function which needs following a long path. Most algorithms fail on this.

    The path becomes thiner as we get closer to the optimum."""
    assert len(x) >= 2
    distance = np.sqrt(x[0] ** 2 + x[1] ** 2)
    if distance == 0.0:
        return 0.0
    angle = np.arctan(x[0] / x[1]) if x[1] != 0.0 else np.pi / 2.0
    invdistance = (1.0 / distance) if distance > 0.0 else 0.0
    if np.abs(np.cos(invdistance) - angle) > 0.1:
        return 1.0
    return float(distance)


@registry.register
def deceptivemultimodal(x: np.ndarray) -> float:
    """Infinitely many local optima, as we get closer to the optimum."""
    assert len(x) >= 2
    distance = np.sqrt(x[0] ** 2 + x[1] ** 2)
    if distance == 0.0:
        return 0.0
    angle = np.arctan(x[0] / x[1]) if x[1] != 0.0 else np.pi / 2.0
    invdistance = int(1.0 / distance) if distance > 0.0 else 0.0
    if np.abs(np.cos(invdistance) - angle) > 0.1:
        return 1.0
    return float(distance)


@registry.register
def lunacek(x: np.ndarray) -> float:
    """Multimodal function.

    Based on https://www.cs.unm.edu/~neal.holts/dga/benchmarkFunction/lunacek.html."""
    problemDimensions = len(x)
    s = 1.0 - (1.0 / (2.0 * np.sqrt(problemDimensions + 20.0) - 8.2))
    mu1 = 2.5
    mu2 = -np.sqrt(abs((mu1**2 - 1.0) / s))
    firstSum = 0.0
    secondSum = 0.0
    thirdSum = 0.0
    for i in range(problemDimensions):
        firstSum += (x[i] - mu1) ** 2
        secondSum += (x[i] - mu2) ** 2
        thirdSum += 1.0 - np.cos(2 * np.pi * (x[i] - mu1))
    return min(firstSum, 1.0 * problemDimensions + secondSum) + 10 * thirdSum


@registry.register_with_info(no_transform=True)
def genzcornerpeak(y: np.ndarray) -> float:
    """One of the Genz functions, originally used in integration,

    tested in optim because why not."""
    value = float(1 + np.mean(np.tanh(y)))
    if value == 0:
        return float("inf")
    return value ** (-len(y) - 1)


@registry.register_with_info(no_transform=True)
def minusgenzcornerpeak(y: np.ndarray) -> float:
    """One of the Genz functions, originally used in integration,

    tested in optim because why not."""
    return -genzcornerpeak(y)


@registry.register
def genzgaussianpeakintegral(x: np.ndarray) -> float:
    """One of the Genz functions, originally used in integration,

    tested in optim because why not."""
    return exp(-sphere(x) / 4.0)


@registry.register
def minusgenzgaussianpeakintegral(x: np.ndarray) -> float:
    """One of the Genz functions, originally used in integration,

    tested in optim because why not."""
    return -genzgaussianpeakintegral(x)


@registry.register
def slope(x: np.ndarray) -> float:
    return sum(x)


@registry.register
def linear(x: np.ndarray) -> float:
    return tanh(x[0])


@registry.register
def st0(x: np.ndarray) -> float:
    """Styblinksitang function with 0 noise."""
    return _styblinksitang(x, 0)


@registry.register
def st1(x: np.ndarray) -> float:
    """Styblinksitang function with noise 1."""
    return _styblinksitang(x, 1)


@registry.register
def st10(x: np.ndarray) -> float:
    """Styblinksitang function with noise 10."""
    return _styblinksitang(x, 10)


@registry.register
def st100(x: np.ndarray) -> float:
    """Styblinksitang function with noise 100."""
    return _styblinksitang(x, 100)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
from nevergrad.common import testing
from . import corefuncs


@testing.parametrized(**{name: (name, func) for name, func in corefuncs.registry.items()})
def testcorefuncs_function(name: str, func: tp.Callable[..., tp.Any]) -> None:
    x = np.random.normal(0, 1, 100)
    outputs = []
    for _ in range(2):
        np.random.seed(12)
        outputs.append(func(x))
    np.testing.assert_equal(outputs[0], outputs[1], f"Function {name} is not deterministic")


@testing.parametrized(
    index1=(1,),
    index2=(2,),
    index3=(3,),
    index4=(4,),
)
def test_bonnans(index: int) -> None:
    b = corefuncs.BonnansFunction(index)
    assert b(np.zeros(100)) < 10.0  # unlikely (though not impossible)
    assert b(np.random.rand(100)) > 0.0


@testing.parametrized(
    expe1=([6, 4, 2, 1, 9], 4, 5, 3),
    expe2=([6, 6, 7, 1, 9], 4, 5, 3),
    expe3=([1, 1, 7, 1, 9], 3, 5, 2),
    expe4=([0, 0, 7, 1, 9], 3, 4, 2),
    expe5=([0, 1, 1, 1], 1, 2, 0),
    expe6=([1, 1, 1, 1], 2, 4, 1),
    expe7=([1, 0, 0, 0], 3, 4, 2),
    expe_0lead=([0, 1, 0, 1], 0, 0, -1),
)  # jump was assumed correct (verify?)
def test_base_functions(
    x: tp.List[float], onemax_expected: float, leadingones_expected: float, jump_expected: float
) -> None:
    for name, expected in [
        ("onemax", onemax_expected),
        ("leadingones", leadingones_expected),
        ("jump", jump_expected),
    ]:
        func = corefuncs.DiscreteFunction(name)
        np.testing.assert_equal(func(x), expected, err_msg=f"Wrong output for {name}")


def test_genzcornerpeak_inf() -> None:
    y = np.array([-np.inf, -np.inf])
    output = corefuncs.genzcornerpeak(y)
    np.testing.assert_equal(output, np.inf)
    output = corefuncs.minusgenzcornerpeak(y)
    np.testing.assert_equal(output, -np.inf)


@testing.parametrized(
    cigar=(corefuncs.cigar, 8604785.43824, None),
    hm=(corefuncs.hm, 15.85037, None),
    griewank=(corefuncs.griewank, 0.70028, None),
    sphere=(corefuncs.sphere, 9.043029, None),
    sphere1=(corefuncs.sphere1, 10.329029, None),
    sphere2=(corefuncs.sphere2, 35.615029, None),
    sphere4=(corefuncs.sphere4, 158.187029, None),
    sphere_m=(corefuncs.sphere, 30, [1, 2, 3, 4]),
    elliptic=(corefuncs.ellipsoid, 262389.541996, None),
    altelliptic=(corefuncs.altellipsoid, 609313.71475, None),
    rosenbrock=(corefuncs.rosenbrock, 1967.57859, None),
    rosenbrock_m=(corefuncs.rosenbrock, 2705, [1, 2, 3, 4]),
    ackley=(corefuncs.ackley, 4.7298473, None),
    schwefel=(corefuncs.schwefel_1_2, 128.709765, None),
    ackley_m=(corefuncs.ackley, 2.254444, [0.1, 0.2, 0.3]),
    schwefel_m=(corefuncs.schwefel_1_2, 146, [1, 2, 3, 4]),
    genzgaussianpeakintegral=(corefuncs.genzgaussianpeakintegral, 0.10427, None),
    minusgenzgaussianpeakintegral=(corefuncs.minusgenzgaussianpeakintegral, -0.10427, None),
    linear=(corefuncs.linear, 0.57969, None),
)
def test_core_function_values(
    func: tp.Callable[[np.ndarray], float], expected: float, data: tp.Optional[tp.List[float]]
) -> None:
    if data is None:
        data = [0.662, -0.217, -0.968, 1.867, 0.101, 0.575, 0.199, 1.576, 1.006, 0.182, -0.092, 0.466]
    value = func(np.array(data))
    np.testing.assert_almost_equal(value, expected, decimal=5)


def test_styblinksitang() -> None:
    np.random.seed(12)
    data = [0.662, -0.217, -0.968, 1.867, 0.101, 0.575, 0.199, 1.576, 1.006, 0.182, -0.092, 0.466]
    value = corefuncs._styblinksitang(np.array(data), noise=0.1)
    np.testing.assert_almost_equal(value, 421.374940, decimal=5)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pickle
import warnings
from pathlib import Path
from numbers import Real
from collections import deque
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p
from nevergrad.common import tools as ngtools
from nevergrad.common import errors as errors
from nevergrad.common.decorators import Registry
from . import utils
from . import multiobjective as mobj


OptCls = tp.Union["ConfiguredOptimizer", tp.Type["Optimizer"]]
registry: Registry[OptCls] = Registry()
_OptimCallBack = tp.Union[
    tp.Callable[["Optimizer", "p.Parameter", float], None], tp.Callable[["Optimizer"], None]
]
X = tp.TypeVar("X", bound="Optimizer")
Y = tp.TypeVar("Y")
IntOrParameter = tp.Union[int, p.Parameter]
_PruningCallable = tp.Callable[[utils.Archive[utils.MultiValue]], utils.Archive[utils.MultiValue]]


def _loss(param: p.Parameter) -> float:
    """Returns the loss if available, or inf otherwise.
    Used to simplify handling of losses
    """
    return param.loss if param.loss is not None else float("inf")


def load(cls: tp.Type[X], filepath: tp.PathLike) -> X:
    """Loads a pickle file and checks that it contains an optimizer.
    The optimizer class is not always fully reliable though (e.g.: optimizer families) so the user is responsible for it.
    """
    filepath = Path(filepath)
    with filepath.open("rb") as f:
        opt = pickle.load(f)
    assert isinstance(opt, cls), f"You should only load {cls} with this method (found {type(opt)})"
    return opt


class Optimizer:  # pylint: disable=too-many-instance-attributes
    """Algorithm framework with 3 main functions:

    - :code:`ask()` which provides a candidate on which to evaluate the function to optimize.
    - :code:`tell(candidate, loss)` which lets you provide the loss associated to points.
    - :code:`provide_recommendation()` which provides the best final candidate.

    Typically, one would call :code:`ask()` num_workers times, evaluate the
    function on these num_workers points in parallel, update with the fitness value when the
    evaluations is finished, and iterate until the budget is over. At the very end,
    one would call provide_recommendation for the estimated optimum.

    This class is abstract, it provides internal equivalents for the 3 main functions,
    among which at least :code:`_internal_ask_candidate` has to be overridden.

    Each optimizer instance should be used only once, with the initial provided budget

    Parameters
    ----------
    parametrization: int or Parameter
        either the dimension of the optimization space, or its parametrization
    budget: int/None
        number of allowed evaluations
    num_workers: int
        number of evaluations which will be run in parallel at once
    """

    # pylint: disable=too-many-locals

    # optimizer qualifiers
    recast = False  # algorithm which were not designed to work with the suggest/update pattern
    one_shot = False  # algorithm designed to suggest all budget points at once
    no_parallelization = False  # algorithm which is designed to run sequentially only

    def __init__(
        self, parametrization: IntOrParameter, budget: tp.Optional[int] = None, num_workers: int = 1
    ) -> None:
        if self.no_parallelization and num_workers > 1:
            raise ValueError(f"{self.__class__.__name__} does not support parallelization")
        # "seedable" random state: externally setting the seed will provide deterministic behavior
        # you can also replace or reinitialize this random state
        self.num_workers = int(num_workers)
        self.budget = budget
        self.optim_curve: tp.List[tp.Any] = []

        # How do we deal with cheap constraints i.e. constraints which are fast and use low resources and easy ?
        # True ==> we penalize them (infinite values for candidates which violate the constraint).
        # False ==> we repeat the ask until we solve the problem.
        self._constraints_manager = utils.ConstraintManager()
        self._penalize_cheap_violations = False

        self.parametrization = (
            parametrization
            if not isinstance(parametrization, (int, np.int_))
            else p.Array(shape=(parametrization,))
        )
        self.parametrization.freeze()  # avoids issues!
        if not self.dimension:
            raise ValueError("No variable to optimize in this parametrization.")
        self.name = self.__class__.__name__  # printed name in repr
        # keep a record of evaluations, and current bests which are updated at each new evaluation
        self.archive: utils.Archive[
            utils.MultiValue
        ] = utils.Archive()  # dict like structure taking np.ndarray as keys and Value as values
        self.current_bests = {
            x: utils.MultiValue(self.parametrization, np.inf, reference=self.parametrization)
            for x in ["optimistic", "pessimistic", "average", "minimum"]
        }
        # pruning function, called at each "tell"
        # this can be desactivated or modified by each implementation
        self.pruning: tp.Optional[_PruningCallable] = utils.Pruning.sensible_default(
            num_workers=num_workers, dimension=self.parametrization.dimension
        )
        # multiobjective
        self._MULTIOBJECTIVE_AUTO_BOUND = mobj.AUTO_BOUND
        self._hypervolume_pareto: tp.Optional[mobj.HypervolumePareto] = None
        # instance state
        self._asked: tp.Set[str] = set()
        self._num_objectives = 0
        self._suggestions: tp.Deque[p.Parameter] = deque()
        self._num_ask = 0
        self._num_tell = 0  # increases after each successful tell
        self._num_tell_not_asked = 0
        self._callbacks: tp.Dict[str, tp.List[tp.Any]] = {}
        # to make optimize function stoppable halway through
        self._running_jobs: tp.List[tp.Tuple[p.Parameter, tp.JobLike[tp.Loss]]] = []
        self._finished_jobs: tp.Deque[tp.Tuple[p.Parameter, tp.JobLike[tp.Loss]]] = deque()
        self._sent_warnings: tp.Set[tp.Any] = set()  # no use printing several time the same
        # Most optimizers are designed for single objective and use a float loss.
        # To use these in a multi-objective optimization, we provide the negative of
        # the hypervolume of the pareto front as the loss.
        # If not needed, an optimizer can set this to True.
        self._no_hypervolume = False

    def _warn(self, msg: str, e: tp.Any) -> None:
        """Warns only once per warning type"""
        if e not in self._sent_warnings:
            warnings.warn(msg, e)
            self._sent_warnings.add(e)

    @property
    def _rng(self) -> np.random.RandomState:
        """np.random.RandomState: parametrization random state the optimizer must pull from.
        It can be seeded or updated directly on the parametrization instance (`optimizer.parametrization.random_state`)
        """
        return self.parametrization.random_state

    @property
    def dimension(self) -> int:
        """int: Dimension of the optimization space."""
        return self.parametrization.dimension

    @property
    def num_objectives(self) -> int:
        """Provides 0 if the number is not known yet, else the number of objectives
        to optimize upon.
        """
        if (
            self._hypervolume_pareto is not None
            and self._num_objectives != self._hypervolume_pareto.num_objectives
        ):
            raise RuntimeError("Number of objectives is incorrectly set. Please create a nevergrad issue")
        return self._num_objectives

    @num_objectives.setter
    def num_objectives(self, num: int) -> None:
        num = int(num)
        if num <= 0:
            raise ValueError("Number of objectives must be strictly positive")
        if not self._num_objectives:
            self._num_objectives = num
            self._num_objectives_set_callback()
        elif num != self._num_objectives:
            raise ValueError(f"Expected {self._num_objectives} loss(es), but received {num}.")

    def _num_objectives_set_callback(self) -> None:
        """Callback for when num objectives is first known"""

    @property
    def num_ask(self) -> int:
        """int: Number of time the `ask` method was called."""
        return self._num_ask

    @property
    def num_tell(self) -> int:
        """int: Number of time the `tell` method was called."""
        return self._num_tell

    @property
    def num_tell_not_asked(self) -> int:
        """int: Number of time the :code:`tell` method was called on candidates that were not asked for by the optimizer
        (or were suggested).
        """
        return self._num_tell_not_asked

    def pareto_front(
        self, size: tp.Optional[int] = None, subset: str = "random", subset_tentatives: int = 12
    ) -> tp.List[p.Parameter]:
        """Pareto front, as a list of Parameter. The losses can be accessed through
        parameter.losses

        Parameters
        ------------
        size:  int (optional)
            if provided, selects a subset of the full pareto front with the given maximum size
        subset: str
            method for selecting the subset ("random, "loss-covering", "domain-covering", "hypervolume")
        subset_tentatives: int
            number of random tentatives for finding a better subset

        Returns
        --------
        list
            the list of Parameter of the pareto front

        Note
        ----
        During non-multiobjective optimization, this returns the current pessimistic best
        """
        pareto = (
            []
            if self._hypervolume_pareto is None
            else self._hypervolume_pareto.pareto_front(
                size=size, subset=subset, subset_tentatives=subset_tentatives
            )
        )
        return pareto if pareto else [self.provide_recommendation()]

    def dump(self, filepath: tp.Union[str, Path]) -> None:
        """Pickles the optimizer into a file."""
        filepath = Path(filepath)
        with filepath.open("wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls: tp.Type[X], filepath: tp.Union[str, Path]) -> X:
        """Loads a pickle and checks that the class is correct."""
        return load(cls, filepath)

    def __repr__(self) -> str:
        inststr = self.parametrization.name
        return f"Instance of {self.name}(parametrization={inststr}, budget={self.budget}, num_workers={self.num_workers})"

    def register_callback(self, name: str, callback: _OptimCallBack) -> None:
        """Add a callback method called either when `tell` or `ask` are called, with the same
        arguments (including the optimizer / self). This can be useful for custom logging.

        Parameters
        ----------
        name: str
            name of the method to register the callback for (either :code:`ask` or :code:`tell`)
        callback: callable
            a callable taking the same parameters as the method it is registered upon (including self)
        """
        assert name in ["ask", "tell"], f'Only "ask" and "tell" methods can have callbacks (not {name})'
        self._callbacks.setdefault(name, []).append(callback)

    def remove_all_callbacks(self) -> None:
        """Removes all registered callables"""
        self._callbacks = {}

    def suggest(self, *args: tp.Any, **kwargs: tp.Any) -> None:
        """Suggests a new point to ask.
        It will be asked at the next call (last in first out).

        Parameters
        ----------
        *args: Any
            positional arguments matching the parametrization pattern.
        *kwargs: Any
            keyword arguments matching the parametrization pattern.

        Note
        ----
        - This relies on optmizers implementing a way to deal with unasked candidate.
          Some optimizers may not support it and will raise a :code:`TellNotAskedNotSupportedError`
          at :code:`tell` time.
        - LIFO is used so as to be able to suggest and ask straightaway, as an alternative to
          creating a new candidate with :code:`optimizer.parametrization.spawn_child(new_value)`
        """
        if isinstance(self.parametrization, p.Instrumentation):
            new_value: tp.Any = (args, kwargs)
        else:
            assert len(args) == 1 and not kwargs
            new_value = args[0]
        self._suggestions.append(self.parametrization.spawn_child(new_value=new_value))

    # pylint: disable=too-many-branches
    def tell(
        self,
        candidate: p.Parameter,
        loss: tp.Loss,
        constraint_violation: tp.Optional[tp.Loss] = None,
        penalty_style: tp.Optional[tp.ArrayLike] = None,
    ) -> None:
        """Provides the optimizer with the evaluation of a fitness value for a candidate.

        Parameters
        ----------
        x: np.ndarray
            point where the function was evaluated
        loss: float/list/np.ndarray
            loss of the function (or multi-objective function
        constraint_violation: float/list/np.ndarray/None
            constraint violation (> 0 means that this is not correct)
        penalty_style: ArrayLike/None
            to be read as [a,b,c,d,e,f]
            with cv the constraint violation vector (above):
            penalty = (a + sum(|loss|)) * (f+num_tell)**e * (b * sum(cv**c)) ** d
            default: [1e5, 1., .5, 1., .5, 1.]

        Note
        ----
        The candidate should generally be one provided by :code:`ask()`, but can be also
        a non-asked candidate. To create a p.Parameter instance from args and kwargs,
        you can use :code:`candidate = optimizer.parametrization.spawn_child(new_value=your_value)`:

        - for an :code:`Array(shape(2,))`: :code:`optimizer.parametrization.spawn_child(new_value=[12, 12])`

        - for an :code:`Instrumentation`: :code:`optimizer.parametrization.spawn_child(new_value=(args, kwargs))`

        Alternatively, you can provide a suggestion with :code:`optimizer.suggest(*args, **kwargs)`, the next :code:`ask`
        will use this suggestion.
        """
        # Check loss type
        if isinstance(loss, (Real, float)) or (isinstance(loss, np.ndarray) and not loss.shape):
            # using "float" along "Real" because mypy does not understand "Real" for now Issue #3186
            loss = float(loss)
            # Non-sense values including NaNs should not be accepted.
            # We do not use max-float as various later transformations could lead to greater values.
            if not loss < 5.0e20:  # pylint: disable=unneeded-not
                self._warn(
                    f"Clipping very high value {loss} in tell (rescale the cost function?).",
                    errors.LossTooLargeWarning,
                )
                loss = 5.0e20  # sys.float_info.max leads to numerical problems so let us do this.
        elif isinstance(loss, (tuple, list, np.ndarray)):
            loss = np.array(loss, copy=False, dtype=float).ravel() if len(loss) != 1 else loss[0]
        elif not isinstance(loss, np.ndarray):
            raise TypeError(
                f'"tell" method only supports float values but the passed loss was: {loss} (type: {type(loss)}.'
            )
        if isinstance(loss, float) and (
            len(self.optim_curve) == 0 or self.num_tell > self.optim_curve[-1][0] * 1.1
        ):
            self.optim_curve += [(self.num_tell, loss)]
        # check Parameter
        if not isinstance(candidate, p.Parameter):
            raise TypeError(
                "'tell' must be provided with the candidate.\n"
                "Use optimizer.parametrization.spawn_child(new_value)) if you want to "
                "create a candidate that as not been asked for, "
                "or optimizer.suggest(*args, **kwargs) to suggest a point that should be used for "
                "the next ask"
            )
        # check loss length
        self.num_objectives = 1 if isinstance(loss, float) else loss.size
        # checks are done, start processing
        candidate.freeze()  # make sure it is not modified somewhere
        # add reference if provided
        if isinstance(candidate, p.MultiobjectiveReference):
            if self._hypervolume_pareto is not None:
                raise RuntimeError("MultiobjectiveReference can only be provided before the first tell.")
            if not isinstance(loss, np.ndarray):
                raise RuntimeError("MultiobjectiveReference must only be used for multiobjective losses")
            self._hypervolume_pareto = mobj.HypervolumePareto(
                upper_bounds=loss, seed=self._rng, no_hypervolume=self._no_hypervolume
            )
            if candidate.value is None:
                return  # no value, so stopping processing there
            candidate = candidate.value
        # preprocess multiobjective loss
        if isinstance(loss, np.ndarray):
            candidate._losses = loss
        if not isinstance(loss, float):
            loss = self._preprocess_multiobjective(candidate)
        # call callbacks for logging etc...
        candidate.loss = loss
        assert isinstance(loss, float)
        for callback in self._callbacks.get("tell", []):
            # multiobjective reference is not handled :s
            # but this allows obtaining both scalar and multiobjective loss (through losses)
            callback(self, candidate, loss)
        if not candidate.satisfies_constraints(self.parametrization) and self.budget is not None:
            penalty = self._constraints_manager.penalty(candidate, self.num_ask, self.budget)
            loss = loss + penalty

        if constraint_violation is not None:
            if penalty_style is not None:
                a, b, c, d, e, f = penalty_style
            else:
                a, b, c, d, e, f = (1e5, 1.0, 0.5, 1.0, 0.5, 1.0)

            violation = float(
                (a + np.sum(np.maximum(loss, 0.0)))
                * ((f + self._num_tell) ** e)
                * (b * np.sum(np.maximum(constraint_violation, 0.0) ** c) ** d)
            )
            loss += violation

        if isinstance(loss, float) and (
            self.num_objectives == 1 or self.num_objectives > 1 and not self._no_hypervolume
        ):
            self._update_archive_and_bests(candidate, loss)

        if candidate.uid in self._asked:
            self._internal_tell_candidate(candidate, loss)
            self._asked.remove(candidate.uid)
        else:
            self._internal_tell_not_asked(candidate, loss)
            self._num_tell_not_asked += 1
        self._num_tell += 1

    def _preprocess_multiobjective(self, candidate: p.Parameter) -> tp.FloatLoss:
        if self._hypervolume_pareto is None:
            self._hypervolume_pareto = mobj.HypervolumePareto(
                auto_bound=self._MULTIOBJECTIVE_AUTO_BOUND, no_hypervolume=self._no_hypervolume
            )
        return self._hypervolume_pareto.add(candidate)

    def _update_archive_and_bests(self, candidate: p.Parameter, loss: tp.FloatLoss) -> None:
        x = candidate.get_standardized_data(reference=self.parametrization)
        if not isinstance(
            loss, (Real, float)
        ):  # using "float" along "Real" because mypy does not understand "Real" for now Issue #3186
            raise TypeError(
                f'"tell" method only supports float values but the passed loss was: {loss} (type: {type(loss)}.'
            )
        if np.isnan(loss) or loss == np.inf:
            self._warn(f"Updating fitness with {loss} value", errors.BadLossWarning)
        mvalue: tp.Optional[utils.MultiValue] = None
        if x not in self.archive:
            self.archive[x] = utils.MultiValue(candidate, loss, reference=self.parametrization)
        else:
            mvalue = self.archive[x]
            mvalue.add_evaluation(loss)
            # both parameters should be non-None
            if mvalue.parameter.loss > candidate.loss:  # type: ignore
                mvalue.parameter = candidate  # keep best candidate
        # update current best records
        # this may have to be improved if we want to keep more kinds of best losss

        for name in self.current_bests:
            if mvalue is self.current_bests[name]:  # reboot
                best = min(self.archive.values(), key=lambda mv, n=name: mv.get_estimation(n))  # type: ignore
                # rebuild best point may change, and which value did not track the updated value anyway
                self.current_bests[name] = best
            else:
                if self.archive[x].get_estimation(name) <= self.current_bests[name].get_estimation(name):
                    self.current_bests[name] = self.archive[x]
                # deactivated checks
                # if not (np.isnan(loss) or loss == np.inf):
                #     if not self.current_bests[name].x in self.archive:
                #         bval = self.current_bests[name].get_estimation(name)
                #         avals = (min(v.get_estimation(name) for v in self.archive.values()),
                #                  max(v.get_estimation(name) for v in self.archive.values()))
                #         raise RuntimeError(f"Best value should exist in the archive at num_tell={self.num_tell})\n"
                #                            f"Best value is {bval} and archive is within range {avals} for {name}")
        if self.pruning is not None:
            self.archive = self.pruning(self.archive)

    def ask(self) -> p.Parameter:
        """Provides a point to explore.
        This function can be called multiple times to explore several points in parallel

        Returns
        -------
        p.Parameter:
            The candidate to try on the objective function. :code:`p.Parameter` have field :code:`args` and :code:`kwargs`
            which can be directly used on the function (:code:`objective_function(*candidate.args, **candidate.kwargs)`).
        """
        # call callbacks for logging etc...
        for callback in self._callbacks.get("ask", []):
            callback(self)
        current_num_ask = self.num_ask
        # tentatives if a cheap constraint is available
        max_trials = max(1, self._constraints_manager.max_trials // 2)
        # half will be used for sub-optimization --- if the optimization method does not need/use a budget.
        # TODO(oteytaud): actually we could do this even when the budget is known, if we are sure that
        # exceeding the budget is not a problem.
        # Very simple constraint solver:
        # - we use a simple algorithm.
        # - no memory of previous iterations.
        # - just projection to constraint satisfaction.
        # We try using the normal tool during half constraint budget, in order to reduce the impact on the normal run.
        self.parametrization.tabu_fails = 0
        for _ in range(max_trials):
            is_suggestion = False
            if self._suggestions:  # use suggestions if available
                is_suggestion = True
                candidate = self._suggestions.pop()
            else:
                try:  # Sometimes we have a limited budget so that
                    candidate = self._internal_ask_candidate()
                except AssertionError as e:
                    assert (
                        self.parametrization._constraint_checkers
                    ), f"Error: {e}"  # This should not happen without constraint issues.
                    candidate = self.parametrization.spawn_child()
            if candidate.satisfies_constraints(self.parametrization):
                break  # good to go!
            if self._penalize_cheap_violations or self.no_parallelization:
                # Warning! This might be a tell not asked.
                self._internal_tell_candidate(candidate, float("Inf"))  # DE requires a tell
            # updating num_ask  is necessary for some algorithms which need new num to ask another point
            self._num_ask += 1
        satisfies = candidate.satisfies_constraints(self.parametrization)
        if not satisfies and self.parametrization.tabu_length == 0:
            # still not solving, let's run sub-optimization
            candidate = _constraint_solver(candidate, budget=max_trials)
        if not (satisfies or candidate.satisfies_constraints(self.parametrization, no_tabu=True)):
            self._warn(
                f"Could not bypass the constraint after {max_trials} tentatives, "
                "sending candidate anyway.",
                errors.FailedConstraintWarning,
            )
        if not is_suggestion:
            if candidate.uid in self._asked:
                raise RuntimeError(
                    "Cannot submit the same candidate twice: please recreate a new candidate from data.\n"
                    "This is to make sure that stochastic parametrizations are resampled."
                )
            self._asked.add(candidate.uid)
        self._num_ask = current_num_ask + 1
        assert (
            candidate is not None
        ), f"{self.__class__.__name__}._internal_ask method returned None instead of a point."
        # make sure to call value getter which may update the value, before we freeze the paremeter
        candidate.value  # pylint: disable=pointless-statement
        candidate.freeze()  # make sure it is not modified somewhere
        return candidate

    def provide_recommendation(self) -> p.Parameter:
        """Provides the best point to use as a minimum, given the budget that was used

        Returns
        -------
        p.Parameter
            The candidate with minimal value. p.Parameters have field :code:`args` and :code:`kwargs` which can be directly used
            on the function (:code:`objective_function(*candidate.args, **candidate.kwargs)`).
        """
        return self.recommend()  # duplicate method

    def recommend(self) -> p.Parameter:
        """Provides the best candidate to use as a minimum, given the budget that was used.

        Returns
        -------
        p.Parameter
            The candidate with minimal loss. :code:`p.Parameters` have field :code:`args` and :code:`kwargs` which can be directly used
            on the function (:code:`objective_function(*candidate.args, **candidate.kwargs)`).
        """
        if self.num_objectives > 1:
            raise RuntimeError(
                "No best candidate in MOO. Use optimizer.pareto_front() instead to get the set of all non-dominated candidates."
            )
        recom_data = self._internal_provide_recommendation()  # pylint: disable=assignment-from-none
        if recom_data is None or any(np.isnan(recom_data)):
            name = "minimum" if self.parametrization.function.deterministic else "pessimistic"
            return self.current_bests[name].parameter
        out = self.parametrization.spawn_child()
        with p.helpers.deterministic_sampling(out):
            out.set_standardized_data(recom_data)
        return out

    def _internal_tell_not_asked(self, candidate: p.Parameter, loss: tp.FloatLoss) -> None:
        """Called whenever calling :code:`tell` on a candidate that was not "asked".
        Defaults to the standard tell pipeline.
        """
        self._internal_tell_candidate(candidate, loss)

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: tp.FloatLoss) -> None:
        """Called whenever calling :code:`tell` on a candidate that was "asked"."""
        data = candidate.get_standardized_data(reference=self.parametrization)
        self._internal_tell(data, loss)

    def _internal_ask_candidate(self) -> p.Parameter:
        return self.parametrization.spawn_child().set_standardized_data(self._internal_ask())

    # Internal methods which can be overloaded (or must be, in the case of _internal_ask)
    def _internal_tell(self, x: tp.ArrayLike, loss: tp.FloatLoss) -> None:
        pass

    def _internal_ask(self) -> tp.ArrayLike:
        raise RuntimeError("Not implemented, should not be called.")

    def _internal_provide_recommendation(self) -> tp.Optional[tp.ArrayLike]:
        """Override to provide a recommendation in standardized space"""
        return None

    def enable_pickling(self) -> None:
        """
        Some optimizers are only optionally picklable, because picklability
        requires saving the whole history which would be a waste of memory
        in general. To tell an optimizer to be picklable, call this function
        before any asks.

        In this base class, the function is a no-op, but it is overridden
        in some optimizers.
        """

    def minimize(
        self,
        objective_function: tp.Callable[..., tp.Loss],
        executor: tp.Optional[tp.ExecutorLike] = None,
        batch_mode: bool = False,
        verbosity: int = 0,
        constraint_violation: tp.Any = None,
    ) -> p.Parameter:
        """Optimization (minimization) procedure

        Parameters
        ----------
        objective_function: callable
            A callable to optimize (minimize)
        executor: Executor
            An executor object, with method :code:`submit(callable, *args, **kwargs)` and returning a Future-like object
            with methods :code:`done() -> bool` and :code:`result() -> float`. The executor role is to dispatch the execution of
            the jobs locally/on a cluster/with multithreading depending on the implementation.
            Eg: :code:`concurrent.futures.ProcessPoolExecutor`
        batch_mode: bool
            when :code:`num_workers = n > 1`, whether jobs are executed by batch (:code:`n` function evaluations are launched,
            we wait for all results and relaunch n evals) or not (whenever an evaluation is finished, we launch
            another one)
        verbosity: int
            print information about the optimization (0: None, 1: fitness values, 2: fitness values and recommendation)
        constraint_violation: list of functions or None
            each function in the list returns >0 for a violated constraint.

        Returns
        -------
        ng.p.Parameter
            The candidate with minimal value. :code:`ng.p.Parameters` have field :code:`args` and :code:`kwargs` which can
            be directly used on the function (:code:`objective_function(*candidate.args, **candidate.kwargs)`).

        Note
        ----
        for evaluation purpose and with the current implementation, it is better to use batch_mode=True
        """
        # pylint: disable=too-many-branches
        if self.budget is None:
            raise ValueError("Budget must be specified")
        if executor is None:
            executor = utils.SequentialExecutor()  # defaults to run everything locally and sequentially
            if self.num_workers > 1:
                self._warn(
                    f"num_workers = {self.num_workers} > 1 is suboptimal when run sequentially",
                    errors.InefficientSettingsWarning,
                )
        assert executor is not None
        tmp_runnings: tp.List[tp.Tuple[p.Parameter, tp.JobLike[tp.Loss]]] = []
        tmp_finished: tp.Deque[tp.Tuple[p.Parameter, tp.JobLike[tp.Loss]]] = deque()
        # go
        sleeper = ngtools.Sleeper()  # manages waiting time depending on execution time of the jobs
        remaining_budget = self.budget - self.num_ask
        first_iteration = True
        #
        while remaining_budget or self._running_jobs or self._finished_jobs:
            # # # # # Update optimizer with finished jobs # # # # #
            # this is the first thing to do when resuming an existing optimization run
            # process finished
            if self._finished_jobs:
                if (remaining_budget or sleeper._start is not None) and not first_iteration:
                    # ignore stop if no more suggestion is sent
                    # this is an ugly hack to avoid warnings at the end of steady mode
                    sleeper.stop_timer()
                while self._finished_jobs:
                    x, job = self._finished_jobs[0]
                    result = job.result()
                    if constraint_violation is not None:
                        self.tell(
                            x, result, [f(x.value) for f in constraint_violation]
                        )  # TODO: this is not parallelized, wtf!
                    else:
                        self.tell(x, result)
                    self._finished_jobs.popleft()  # remove it after the tell to make sure it was indeed "told" (in case of interruption)
                    if verbosity:
                        print(f"Updating fitness with value {job.result()}")
                if verbosity:
                    print(f"{remaining_budget} remaining budget and {len(self._running_jobs)} running jobs")
                    if verbosity > 1:
                        print("Current pessimistic best is: {}".format(self.current_bests["pessimistic"]))
            elif not first_iteration:
                sleeper.sleep()
            # # # # # Start new jobs # # # # #
            if not batch_mode or not self._running_jobs:
                new_sugg = max(0, min(remaining_budget, self.num_workers - len(self._running_jobs)))
                if verbosity and new_sugg:
                    print(f"Launching {new_sugg} jobs with new suggestions")
                for _ in range(new_sugg):
                    try:
                        args = self.ask()
                    except errors.NevergradEarlyStopping:
                        remaining_budget = 0
                        break
                    self._running_jobs.append(
                        (args, executor.submit(objective_function, *args.args, **args.kwargs))
                    )
                if new_sugg:
                    sleeper.start_timer()
            if remaining_budget > 0:  # early stopping sets it to 0
                remaining_budget = self.budget - self.num_ask
            # split (repopulate finished and runnings in only one loop to avoid
            # weird effects if job finishes in between two list comprehensions)
            tmp_runnings, tmp_finished = [], deque()
            for x_job in self._running_jobs:
                (tmp_finished if x_job[1].done() else tmp_runnings).append(x_job)
            self._running_jobs, self._finished_jobs = tmp_runnings, tmp_finished
            first_iteration = False
        return self.provide_recommendation() if self.num_objectives == 1 else p.Constant(None)

    def _info(self) -> tp.Dict[str, tp.Any]:
        """Easy access to debug/benchmark info"""
        return {}


# Adding a comparison-only functionality to an optimizer.
def addCompare(optimizer: Optimizer) -> None:
    def compare(self: Optimizer, winners: tp.List[p.Parameter], losers: tp.List[p.Parameter]) -> None:
        # This means that for any i and j, winners[i] is better than winners[i+1], and better than losers[j].
        # This is for cases in which we do not know fitness values, we just know comparisons.

        ref = self.parametrization
        # Evaluate the best fitness value among losers.
        best_fitness_value = 0.0
        for candidate in losers:
            data = candidate.get_standardized_data(reference=self.parametrization)
            if data in self.archive:
                best_fitness_value = min(best_fitness_value, self.archive[data].get_estimation("average"))

        # Now let us decide the fitness value of winners.
        for i, candidate in enumerate(winners):
            self.tell(candidate, best_fitness_value - len(winners) + i)
            data = candidate.get_standardized_data(reference=self.parametrization)
            self.archive[data] = utils.MultiValue(
                candidate, best_fitness_value - len(winners) + i, reference=ref
            )

    setattr(optimizer.__class__, "compare", compare)


class ConfiguredOptimizer:
    """Creates optimizer-like instances with configuration.

    Parameters
    ----------
    OptimizerClass: type
        class of the optimizer to configure, or another ConfiguredOptimizer (config will then be ignored
        except for the optimizer name/representation)
    config: dict
        dictionnary of all the configurations
    as_config: bool
        whether to provide all config as kwargs to the optimizer instantiation (default, see ConfiguredCMA for an example),
        or through a config kwarg referencing self. (if True, see EvolutionStrategy for an example)

    Note
    ----
    This provides a default repr which can be bypassed through set_name
    """

    # optimizer qualifiers
    recast = False  # algorithm which were not designed to work with the suggest/update pattern
    one_shot = False  # algorithm designed to suggest all budget points at once
    no_parallelization = False  # algorithm which is designed to run sequentially only

    def __init__(self, OptimizerClass: OptCls, config: tp.Dict[str, tp.Any], as_config: bool = False) -> None:
        self._OptimizerClass = OptimizerClass
        config.pop("self", None)  # self comes from "locals()"
        config.pop("__class__", None)  # self comes from "locals()"
        self._as_config = as_config
        self._config = config  # keep all, to avoid weird behavior at mismatch between optim and configoptim
        diff = ngtools.different_from_defaults(instance=self, instance_dict=config, check_mismatches=True)
        params = ", ".join(f"{x}={y!r}" for x, y in sorted(diff.items()))
        self.name = f"{self.__class__.__name__}({params})"
        if not as_config:
            # try instantiating for init checks
            # if as_config: check can be done before setting attributes
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", category=errors.InefficientSettingsWarning
                )  # this check does not need to be efficient
                self(parametrization=4, budget=100)

    def config(self) -> tp.Dict[str, tp.Any]:
        return dict(self._config)

    def __call__(
        self, parametrization: IntOrParameter, budget: tp.Optional[int] = None, num_workers: int = 1
    ) -> Optimizer:
        """Creates an optimizer from the parametrization

        Parameters
        ----------
        instrumentation: int or Instrumentation
            either the dimension of the optimization space, or its instrumentation
        budget: int/None
            number of allowed evaluations
        num_workers: int
            number of evaluations which will be run in parallel at once
        """
        config = dict(config=self) if self._as_config else self.config()
        if isinstance(self._OptimizerClass, ConfiguredOptimizer):
            config = {}  # ignore, it's already configured
        run = self._OptimizerClass(parametrization=parametrization, budget=budget, num_workers=num_workers, **config)  # type: ignore
        run.name = self.name
        # hacky but convenient to have around:
        run._configured_optimizer = self  # type: ignore
        return run

    def __repr__(self) -> str:
        return self.name

    def set_name(self, name: str, register: bool = False) -> "ConfiguredOptimizer":
        """Set a new representation for the instance"""
        self.name = name
        if register:
            registry.register_name(name, self)
        return self

    def load(self, filepath: tp.Union[str, Path]) -> "Optimizer":
        """Loads a pickle and checks that it is an Optimizer."""
        return self._OptimizerClass.load(filepath)

    def __eq__(self, other: tp.Any) -> tp.Any:
        if self.__class__ == other.__class__:
            if self._config == other._config:
                return True
        return False


def _constraint_solver(parameter: p.Parameter, budget: int) -> p.Parameter:
    """Runs a suboptimization to solve the parameter constraints"""
    parameter_without_constraint = parameter.copy()
    parameter_without_constraint._constraint_checkers.clear()
    parameter_without_constraint.tabu_length = 0
    opt = registry["OnePlusOne"](parameter_without_constraint, num_workers=1, budget=budget)
    for _ in range(budget):
        cand = opt.ask()
        # Our objective function is minimum for the point the closest to
        # the original candidate under the constraints.
        penalty = sum(utils._float_penalty(func(cand.value)) for func in parameter._constraint_checkers)

        # TODO: this may not scale well with dimension
        distance = np.tanh(np.sum(cand.get_standardized_data(reference=parameter) ** 2))
        # TODO: because of the return whenever constraints are satisfied, the first case never arises
        loss = distance if penalty <= 0 else penalty + distance + 1.0
        opt.tell(cand, loss)
        if penalty <= 0:  # constraints are satisfied
            break
    data = opt.recommend().get_standardized_data(reference=parameter_without_constraint)
    return parameter.spawn_child().set_standardized_data(data)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import itertools
import typing as tp
import numpy as np
from nevergrad.parametrization import parameter as p
from .base import ExperimentFunction
from . import corefuncs


class PBT(ExperimentFunction):
    """Population-Based Training, also known as Lamarckism or Meta-Optimization."""

    def __init__(
        self,
        names: tp.Tuple[str, ...] = ("sphere", "cigar", "ellipsoid"),
        dimensions: tp.Tuple[int, ...] = (7, 7, 7),
        num_workers: int = 10,
    ):
        for name in names:
            if name not in corefuncs.registry:
                available = ", ".join(sorted(corefuncs.registry))
                raise ValueError(
                    f'Unknown core function "{name}" in PBT. Available names are:\n-----\n{available}'
                )
        self._funcs = [corefuncs.registry[name] for name in names]
        self._optima = [np.random.normal(size=d) for d in dimensions]
        assert len(names) == len(dimensions)
        self._hyperparameter_dimension = len(names)
        self._dimensions = dimensions
        self._total_dimension = sum(dimensions)
        parametrization = p.Array(shape=(self._hyperparameter_dimension,)).set_name("")
        # Population of checkpoints (that are optimized by the underlying optimization method)
        # and parameters (that we do optimize).
        self._population_checkpoints: tp.List[np.ndarray] = [np.zeros(self._total_dimension)] * num_workers
        self._population_parameters: tp.List[np.ndarray] = [
            np.zeros(self._hyperparameter_dimension)
        ] * num_workers
        self._population_fitness: tp.List[float] = [float("inf")] * num_workers
        super().__init__(self._func, parametrization)

    # The 3 methods below are function-specific.
    def unflatten(self, x):
        y = []
        current_idx = 0
        for i in range(self._hyperparameter_dimension):
            y += [x[current_idx : (current_idx + self._dimensions[i])]]
            current_idx += self._dimensions[i]
        assert current_idx == len(x) == self._total_dimension
        return y

    def value(self, x):
        y = self.unflatten(x)
        return sum(f(xi - o) for f, xi, o in zip(self._funcs, y, self._optima))

    def evolve(self, x: np.ndarray, pp: np.ndarray):
        assert len(pp) == self._hyperparameter_dimension

        # We define a gradient operator.
        def gradient(f, x):
            epsilon = 1e-15
            # We compute a gradient by finite differences.
            g = np.zeros(len(x))
            value_minus = f(x)
            assert isinstance(value_minus, float), str(type(value_minus))
            for i in range(len(x)):
                e = np.zeros(len(x))
                e[i] = epsilon
                value_plus = f(x + e)
                assert isinstance(value_plus, float), str(type(value_plus))
                g[i] = (value_plus - value_minus) / epsilon
            return g

        y = self.unflatten(x)
        assert len(y) == self._hyperparameter_dimension

        # There are self._hyperparameter_dimension learning rates.
        # Each of them is responsible of one gradient descent.
        for j in range(self._hyperparameter_dimension):
            assert len(y[j]) == self._dimensions[j]
            assert isinstance(self._funcs[j](y[j]), float), str(type(self._funcs[j](y[j])))
            y[j] -= np.exp(pp[j]) * (
                gradient(self._funcs[j], y[j] - self._optima[j]) + np.random.normal(self._dimensions[j])
            )
        current_idx = 0
        for j in range(self._hyperparameter_dimension):
            x[current_idx : (current_idx) + self._dimensions[j]] = y[j]
            current_idx += self._dimensions[j]
        assert current_idx == self._total_dimension

    def _func(self, x: np.ndarray):
        assert len(x) == self._hyperparameter_dimension

        # First, let us find the checkpoint that we want to use.
        if np.random.uniform() > 0.666:
            distances = self._population_fitness
        else:
            if np.random.uniform() > 0.5:
                distances = [np.linalg.norm(i - x, 0) for i in self._population_parameters]  # type: ignore
            else:
                distances = [np.linalg.norm(i - x, 1) for i in self._population_parameters]  # type: ignore
        _, source_idx = min((val, idx) for (idx, val) in enumerate(distances))

        # Let us copy the checkpoint to a target.
        idx = np.random.choice(range(len(self._population_fitness)))
        if idx != source_idx:
            self._population_checkpoints[idx] = self._population_checkpoints[source_idx].copy()
            self._population_fitness[idx] = self._population_fitness[source_idx]
        self._population_parameters[idx] = x

        # Here the case-specific learning and evaluation.
        self.evolve(self._population_checkpoints[idx], self._population_parameters[idx])
        self._population_fitness[idx] = self.value(self._population_checkpoints[idx])

        return self._population_fitness[idx]

    @classmethod
    def itercases(cls) -> tp.Iterator["PBT"]:
        options = dict(
            names=[["sphere", "cigar", "ellipsoid", "cigar"], ["hm", "rastrigin", "sphere", "cigar"]],
            dimensions=[[2, 2, 2, 2], [5, 5, 5, 5], [77, 77, 77, 77]],
        )
        keys = sorted(options)
        select = itertools.product(*(options[k] for k in keys))  # type: ignore
        cases = (dict(zip(keys, s)) for s in select)
        return (cls(**c) for c in cases)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest
import numpy as np
from .core import BaseFunction


class MountainCarContinuous(BaseFunction):
    env_name = "MountainCarContinuous-v0"
    policy_dim = (2, 1)
    state_mean = [0, 0]
    state_std = [1, 1]


@pytest.mark.parametrize("intermediate_layer_dim", [None, (4,), (3, 3)])
@pytest.mark.parametrize("noise_level", [0.0, 0.9])
@pytest.mark.parametrize("deterministic_sim", [0.0, 0.9])
@pytest.mark.parametrize("states_normalization", [True, False])
def test_gym(intermediate_layer_dim, noise_level, deterministic_sim, states_normalization) -> None:
    func = MountainCarContinuous(
        num_rollouts=2,
        intermediate_layer_dim=intermediate_layer_dim,
        noise_level=noise_level,
        deterministic_sim=deterministic_sim,
        states_normalization=states_normalization,
        random_state=42,
    )

    x = func.parametrization.value
    value = func(x)
    np.testing.assert_almost_equal(value, 0.0)
    x = func.parametrization.sample().value
    assert func(x) != 0
    if noise_level == 0.0 and deterministic_sim:
        assert func(x) == func(x)
    else:
        assert func(x) != func(x)


@pytest.mark.parametrize("module", ["Ant", "Swimmer", "HalfCheetah", "Hopper", "Walker2d", "Humanoid"])
@pytest.mark.parametrize("intermediate_layer_dim", [None, (50,)])
@pytest.mark.parametrize("noise_level", [0.0, 0.9])
@pytest.mark.parametrize("deterministic_sim", [0.0, 0.9])
@pytest.mark.parametrize("states_normalization", [True, False])
def test_all_mujoco_envs(
    module, intermediate_layer_dim, noise_level, deterministic_sim, states_normalization
) -> None:
    pytest.importorskip("mujoco_py")
    core = pytest.importorskip("nevergrad.functions.control.core")
    func = getattr(core, module)(
        num_rollouts=1,
        intermediate_layer_dim=intermediate_layer_dim,
        noise_level=noise_level,
        deterministic_sim=deterministic_sim,
        states_normalization=states_normalization,
        random_state=42,
    )
    for _ in range(3):
        assert isinstance(func(func.parametrization.sample().value), float)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import os
import numpy as np
import pytest
import nevergrad.common.typing as tp
from . import test_optimizerlib
from . import optimizerlib as optlib
import nevergrad as ng


KEY = "NEVERGRAD_SPECIAL_TESTS"
if not os.environ.get(KEY, ""):
    pytest.skip(f"These tests only run if {KEY} is set in the environment", allow_module_level=True)


class SimpleFitness:
    """Simple quadratic fitness function which can be used with dimension up to 4"""

    def __init__(self, x0: tp.ArrayLike, x1: tp.ArrayLike) -> None:
        self.x0 = np.array(x0, copy=True)
        self.x1 = np.array(np.exp(x1), copy=True)

    def __call__(self, x: tp.ArrayLike) -> float:
        assert len(self.x0) == len(x)
        return float(np.sum(self.x1 * np.cos(np.array(x, copy=False) - self.x0) ** 2))


@pytest.mark.parametrize("dim", [2, 10, 40, 200])  # type: ignore
@pytest.mark.parametrize("bounded", [True])  # type: ignore
@pytest.mark.parametrize("discrete", [False])  # type: ignore
def test_performance_ngopt(dim: int, bounded: bool, discrete: bool) -> None:
    instrumentation = ng.p.Array(shape=(dim,))
    if dim > 40 and not discrete or dim <= 40 and discrete:
        return
    if bounded:
        instrumentation.set_bounds(lower=-12.0, upper=15.0)
    if discrete:
        instrumentation.set_integer_casting()
    algorithms = [optlib.NGOpt, optlib.OnePlusOne, optlib.CMA]
    if discrete:
        algorithms = [optlib.NGOpt, optlib.DiscreteOnePlusOne, optlib.DoubleFastGADiscreteOnePlusOne]
    num_tests = 7
    fitness = []
    for i in range(num_tests):
        target = np.random.normal(0.0, 1.0, size=dim)
        factors = np.random.normal(0.0, 7.0, size=dim)
        fitness += [SimpleFitness(target, factors)]
    result_tab = []
    for alg in algorithms:
        results = []
        for i in range(num_tests):
            result_for_this_fitness = []
            for budget_multiplier in [10, 100, 1000]:
                for num_workers in [1, 20]:
                    opt = alg(  # type: ignore
                        ng.p.Array(shape=(dim,)), budget=budget_multiplier * dim, num_workers=num_workers
                    )
                    recom = opt.minimize(fitness[i])
                    result_for_this_fitness += [fitness[i](recom.value)]
            results += result_for_this_fitness
        result_tab += [results]
        won_comparisons = [r < ngopt_res for (r, ngopt_res) in zip(result_tab[-1], result_tab[0])]
        assert (
            sum(won_comparisons) < len(won_comparisons) // 2
        ), f"alg={alg}, dim={dim}, budget_multiplier={budget_multiplier}, num_workers={num_workers}, bounded={bounded}, discrete={discrete}, result = {result_tab}"


@pytest.mark.parametrize("args", test_optimizerlib.get_metamodel_test_settings(special=True))
@pytest.mark.parametrize("baseline", ("CMA", "ECMA"))
def test_metamodel_special(baseline: str, args: tp.Tuple[tp.Any, ...]) -> None:
    """The test can operate on the sphere or on an elliptic funciton."""
    kwargs = dict(zip(test_optimizerlib.META_TEST_ARGS, args))
    test_optimizerlib.check_metamodel(baseline=baseline, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import uuid
import warnings
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import errors
from . import utils
from ._layering import ValueProperty as ValueProperty
from ._layering import Layered as Layered
from ._layering import Level as Level


# pylint: disable=no-value-for-parameter,pointless-statement,import-outside-toplevel


P = tp.TypeVar("P", bound="Parameter")


def default_congruence(x: tp.Any) -> tp.Any:
    return x


# pylint: disable=too-many-public-methods
class Parameter(Layered):
    """Class providing the core functionality of a parameter, aka
    value, internal/model parameters, mutation, recombination
    and additional features such as shared random state,
    constraint check, hashes, generation and naming.
    The value field should sent to the function to optimize.

    Example
    -------
    >>> ng.p.Array(shape=(2,)).value
    array([0., 0.])
    """

    # By default, all Parameter attributes of this Parameter are considered as
    # sub-parameters.
    # Spawning a child creates a shallow copy.

    _LAYER_LEVEL = Level.ROOT
    value: ValueProperty[tp.Any, tp.Any] = ValueProperty()

    def __init__(self) -> None:
        # Main features
        super().__init__()
        self.tabu_congruence: tp.Any = default_congruence
        self.tabu_fails = 0
        self._subobjects = utils.Subobjects(
            self, base=Parameter, attribute="__dict__"
        )  # registers and apply functions too all (sub-)Parameter attributes
        self.tabu_length = 0
        self.tabu_set: tp.Set[tp.Any] = set()
        self.tabu_list: tp.List[tp.Any] = []
        self.tabu_index = 0
        self.parents_uids: tp.List[str] = []
        self.heritage: tp.Dict[tp.Hashable, tp.Any] = {"lineage": self.uid}  # passed through to children
        self.loss: tp.Optional[float] = None  # associated loss
        self._losses: tp.Optional[np.ndarray] = None  # associated losses (multiobjective) as an array
        self._dimension: tp.Optional[int] = None
        # Additional convenient features
        self._random_state: tp.Optional[np.random.RandomState] = None  # lazy initialization
        self._generation = 0
        self._constraint_checkers: tp.List[tp.Callable[[tp.Any], tp.Union[bool, float]]] = []
        self._name: tp.Optional[str] = None
        self._frozen = False
        self._meta: tp.Dict[tp.Hashable, tp.Any] = {}  # for anything algorithm related
        self.function = utils.FunctionInfo()

    @property
    def losses(self) -> np.ndarray:
        """Possibly multiobjective losses which were told
        to the optimizer along this parameter.
        In case of mono-objective loss, losses is the array containing this loss as sole element

        Note
        ----
        This API is highly experimental
        """
        if self._losses is not None:
            return self._losses
        if self.loss is not None:
            return np.array([self.loss], dtype=float)
        raise RuntimeError("No loss was provided")

    @property
    def args(self) -> tp.Tuple[tp.Any, ...]:
        """Value of the positional arguments.
        Used to input value in a function as `func(*param.args, **param.kwargs)`
        Use `parameter.Instrumentation` to set `args` and `kwargs` with full freedom.
        """
        return (self.value,)

    @property
    def kwargs(self) -> tp.Dict[str, tp.Any]:
        """Value of the keyword arguments.
        Used to input value in a function as `func(*param.args, **param.kwargs)`
        Use `parameter.Instrumentation` to set `args` and `kwargs` with full freedom.
        """
        return {}

    @property
    def dimension(self) -> int:
        """Dimension of the standardized space for this parameter
        i.e size of the vector returned by get_standardized_data(reference=...)
        """
        if self._dimension is None:
            try:
                self._dimension = self.get_standardized_data(reference=self).size
            except errors.UnsupportedParameterOperationError:
                self._dimension = 0
        return self._dimension

    def mutate(self) -> None:
        """Mutate parameters of the instance, and then its value"""
        self._check_frozen()
        self._subobjects.apply("mutate")
        self._layers[-1]._layered_mutate()

    def _layered_mutate(self) -> None:
        self.set_standardized_data(self.random_state.normal(size=self.dimension))

    def sample(self: P) -> P:
        """Sample a new instance of the parameter.
        This usually means spawning a child and mutating it.
        This function should be used in optimizers when creating an initial population,
        and parameter.heritage["lineage"] is reset to parameter.uid instead of its parent's
        """
        # inner working can be overrided by _layer_sample()
        self.random_state  # make sure to populate it before copy
        child = self._layers[-1]._layered_sample()
        if not isinstance(child, Parameter) and not isinstance(child, type(self)):
            raise errors.NevergradRuntimeError("Unexpected sample return type")
        child._set_parenthood(None)
        return child  # type: ignore

    def recombine(self: P, *others: P) -> None:
        """Update value and parameters of this instance by combining it with
        other instances.

        Parameters
        ----------
        *others: Parameter
            other instances of the same type than this instance.
        """
        if not others:
            return
        self._check_frozen()
        assert all(isinstance(o, self.__class__) for o in others)
        self._subobjects.apply("recombine", *others)
        self._layers[-1]._layered_recombine(*others)

    def get_standardized_data(self: P, *, reference: P) -> np.ndarray:
        """Get the standardized data representing the value of the instance as an array in the optimization space.
        In this standardized space, a mutation is typically centered and reduced (sigma=1) Gaussian noise.
        The data only represent the value of this instance, not the parameters (eg.: mutable sigma), hence it does not
        fully represent the state of the instance. Also, in stochastic cases, the value can be non-deterministically
        deduced from the data (eg.: categorical variable, for which data includes sampling weights for each value)

        Parameters
        ----------
        reference: Parameter
            the reference instance for representation in the standardized data space. This keyword parameter is
            mandatory to make the code clearer.
            If you use "self", this method will always return a zero vector.

        Returns
        -------
        np.ndarray
            the representation of the value in the optimization space

        Note
        ----
        - Operations between different standardized data should only be performed if each array was produced
          by the same reference in the exact same state (no mutation)
        - to make the code more explicit, the "reference" parameter is enforced as a keyword-only parameter.
        """
        assert reference is None or isinstance(
            reference, self.__class__
        ), f"Expected {type(self)} but got {type(reference)} as reference"
        return self._internal_get_standardized_data(self if reference is None else reference)

    def _internal_get_standardized_data(self: P, reference: P) -> np.ndarray:
        raise errors.UnsupportedParameterOperationError(
            f"Export to standardized data space is not implemented for {self.name}"
        )

    def set_standardized_data(self: P, data: tp.ArrayLike, *, reference: tp.Optional[P] = None) -> P:
        """Updates the value of the provided reference (or self) using the standardized data.

        Parameters
        ----------
        np.ndarray
            the representation of the value in the optimization space
        reference: Parameter
            the reference point for representing the data ("self", if not provided)

        Returns
        -------
        Parameter
            self (modified)

        Note
        ----
        To make the code more explicit, the "reference" is enforced
        as keyword-only parameters.
        """
        sent_reference = self if reference is None else reference
        assert isinstance(
            sent_reference, self.__class__
        ), f"Expected {type(self)} but got {type(sent_reference)} as reference"
        self._check_frozen()
        del self.value  # remove all cached information
        self._internal_set_standardized_data(np.array(data, copy=False), reference=sent_reference)
        return self

    def _internal_set_standardized_data(  # pylint: disable=unused-argument
        self: P, data: np.ndarray, reference: P
    ) -> None:
        if data.size:
            raise errors.UnsupportedParameterOperationError(
                f"Import from standardized data space is not implemented for {self.name}"
            )

    # PART 2 - Additional features

    @property
    def generation(self) -> int:
        """Generation of the parameter (children are current generation + 1)"""
        return self._generation

    def get_value_hash(self) -> tp.Hashable:
        """Hashable object representing the current value of the instance"""
        val = self.value
        if isinstance(val, (str, bytes, float, int)):
            return val
        elif isinstance(val, np.ndarray):
            return val.tobytes()
        else:
            raise errors.UnsupportedParameterOperationError(
                f"Value hash is not supported for object {self.name}"
            )

    def __repr__(self) -> str:
        strings = [self.name]
        if not callable(self.value):  # not a mutation
            strings.append(str(self.value))
        return ":".join(strings)

    def __bool__(self) -> bool:
        raise RuntimeError("bool check is not allowed to avoid confusion")

    # %% Constraint management
    def satisfies_constraints(self, ref: tp.Optional[P] = None, no_tabu: bool = False) -> bool:
        """Whether the instance satisfies the constraints added through
        the `register_cheap_constraint` method

        Parameters
        ----------
        ref
            parameter of the optimization algorithm, if we want to check the tabu list.
        no_tabu
            if we want to ignore the Tabu system.
        Returns
        -------
        bool
            True iff the constraint is satisfied
        """
        inside = self._subobjects.apply("satisfies_constraints")
        if not all(inside.values()):
            return False
        val = self.value
        if not no_tabu and ref is not None and ref.tabu_length > 0 and self.tabu_fails < 30:
            tabu_val = self.tabu_congruence(val)
            if isinstance(tabu_val, np.ndarray):
                tabu_val = hash(tabu_val.tobytes())
            if isinstance(tabu_val, dict):
                keys = sorted(list(tabu_val.keys()))
                tabu_val = str(keys) + str([tabu_val[k] for k in keys])
            try:
                if tabu_val in ref.tabu_set:
                    self.tabu_fails += 1
                    return False
                else:
                    ref.tabu_set.add(tabu_val)
                    if len(ref.tabu_list) > ref.tabu_index:
                        ref.tabu_set.remove(ref.tabu_list[ref.tabu_index])
                        ref.tabu_list[ref.tabu_index] = tabu_val
                    else:
                        assert len(ref.tabu_list) == ref.tabu_index
                        ref.tabu_list += [tabu_val]
                    ref.tabu_list[ref.tabu_index] = tabu_val
                    ref.tabu_index = (ref.tabu_index + 1) % ref.tabu_length
            except RuntimeError as e:
                raise RuntimeError(f"{tabu_val} has type {type(tabu_val)}, and this leads to {e}.")
        if not self._constraint_checkers:
            return True
        return all(utils.float_penalty(func(val)) <= 0 for func in self._constraint_checkers)

    def specify_tabu_length(
        self,
        tabu_length: int,
    ) -> None:
        self.tabu_length = tabu_length

    def register_cheap_constraint(
        self,
        func: tp.Union[tp.Callable[[tp.Any], bool], tp.Callable[[tp.Any], float]],
        as_layer: bool = False,
    ) -> None:
        """Registers a new constraint on the parameter values.

        Parameters
        ----------
        func: Callable
            function which, given the value of the instance, returns whether it satisfies the constraints (if output = bool),
            or a float which is >= 0 if the constraint is satisfied.

        Note
        ----
        - this is only for checking after mutation/recombination/etc if the value still satisfy the constraints.
          The constraint is not used in those processes.
        - constraints should be fast to compute.
        - this function has an additional "as_layer" parameter which is experimental for now, and can have unexpected
          behavior
        """
        if getattr(func, "__name__", "not lambda") == "<lambda>":  # LambdaType does not work :(
            warnings.warn("Lambda as constraint is not advised because it may not be picklable.")
        if not as_layer:
            self._constraint_checkers.append(func)
        else:
            from nevergrad.ops.constraints import Constraint
            import nevergrad as ng

            compat_func = (
                func
                if not isinstance(self, ng.p.Instrumentation)
                else utils._ConstraintCompatibilityFunction(func)
            )
            self.add_layer(Constraint(compat_func))  # type: ignore

    # %% random state

    @property
    def random_state(self) -> np.random.RandomState:
        """Random state the instrumentation and the optimizers pull from.
        It can be seeded/replaced.
        """
        if self._random_state is None:
            # use the setter, to make sure the random state is propagated to the variables
            seed = np.random.randint(2**32, dtype=np.uint32)  # better way?
            self._set_random_state(np.random.RandomState(seed))
        assert self._random_state is not None
        return self._random_state

    @random_state.setter
    def random_state(self, random_state: tp.Optional[np.random.RandomState]) -> None:
        self._set_random_state(random_state)

    def _set_random_state(self, random_state: tp.Optional[np.random.RandomState]) -> None:
        self._random_state = random_state
        self._subobjects.apply("_set_random_state", random_state)

    def spawn_child(self: P, new_value: tp.Optional[tp.Any] = None) -> P:
        """Creates a new instance which shares the same random generator than its parent,
        is sampled from the same data, and mutates independently from the parentp.
        If a new value is provided, it will be set to the new instance

        Parameters
        ----------
        new_value: anything (optional)
            if provided, it will update the new instance value (cannot be used at the same time as new_data).

        Returns
        -------
        Parameter
            a new instance of the same class, with same content/internal-model parameters/...
            Optionally, a new value will be set after creation
        """
        # make sure to initialize the random state  before spawning children
        self.random_state  # pylint: disable=pointless-statement
        child = self.copy()
        child._set_parenthood(self)
        if new_value is not None:
            child.value = new_value
        return child

    def copy(self: P) -> P:
        """Creates a full copy of the parameter (with new unique uid).
        Use spawn_child instead to make sure to add the parenthood information.
        """
        child = super().copy()
        child.uid = uuid.uuid4().hex
        child._frozen = False
        child._subobjects = self._subobjects.new(child)
        child._meta = {}
        child.tabu_length = self.tabu_length
        child.parents_uids = list(self.parents_uids)
        child.heritage = dict(self.heritage)
        child.loss = None
        child._losses = None
        child._constraint_checkers = list(self._constraint_checkers)
        # layers
        if self is not self._layers[0]:
            raise errors.NevergradRuntimeError("Something has gone horribly wrong with the layers")
        # subparameters
        attribute = self._subobjects.attribute
        container = getattr(child, attribute)
        if attribute != "__dict__":  # make a copy of the container if different from __dict__
            container = dict(container) if isinstance(container, dict) else list(container)
            setattr(child, attribute, container)
        for key, val in self._subobjects.items():
            container[key] = val.copy()
        del child.value  # clear cache
        return child

    def _set_parenthood(self, parent: tp.Optional["Parameter"]) -> None:
        """Sets the parenthood information to Parameter and subparameters."""
        if parent is None:
            self._generation = 0
            self.heritage = dict(lineage=self.uid)
            self.parents_uids = []
        else:
            self._generation = parent.generation + 1
            self.parents_uids = [parent.uid]
        self._subobjects.apply("_set_parenthood", parent)

    def freeze(self) -> None:
        """Prevents the parameter from changing value again (through value, mutate etc...)"""
        self._frozen = True
        self._subobjects.apply("freeze")

    def _check_frozen(self) -> None:
        if self._frozen and not isinstance(
            self, Constant
        ):  # nevermind constants (since they dont spawn children)
            raise RuntimeError(
                f"Cannot modify frozen Parameter {self.name}, please spawn a child and modify it instead"
                "(optimizers freeze the parametrization and all asked and told candidates to avoid border effects)"
            )
        # make sure the random state is initialized if we need to update it (aka if not frozen)
        self.random_state  # pylint: disable=pointless-statement
        self._subobjects.apply("_check_frozen")


# Basic types #
class Constant(Parameter):
    """Parameter-like object for simplifying management of constant parameters:
    mutation/recombination do nothing, value cannot be changed, standardize data is an empty array,
    child is the same instance.

    Parameter
    ---------
    value: Any
        the value that this parameter will always provide
    """

    def __init__(self, value: tp.Any) -> None:
        super().__init__()
        if isinstance(value, Parameter) and not isinstance(self, MultiobjectiveReference):
            raise TypeError("Only non-parameters can be wrapped in a Constant")
        self._value = value

    def _get_name(self) -> str:
        return str(self._value)

    def get_value_hash(self) -> tp.Hashable:
        try:
            return super().get_value_hash()
        except errors.UnsupportedParameterOperationError:
            return "#non-hashable-constant#"

    def _layered_get_value(self) -> tp.Any:
        return self._value

    def _layered_set_value(self, value: tp.Any) -> None:
        different = False
        if isinstance(value, np.ndarray):
            if not np.equal(value, self._value).all():
                different = True
        elif not (value == self._value or value is self._value):
            different = True
        if different:
            raise ValueError(
                f'Constant value can only be updated to the same value (in this case "{self._value}")'
            )

    def _layered_sample(self: P) -> P:
        return self

    def get_standardized_data(  # pylint: disable=unused-argument
        self: P, *, reference: tp.Optional[P] = None
    ) -> np.ndarray:
        return np.array([])

    def spawn_child(self: P, new_value: tp.Optional[tp.Any] = None) -> P:
        if new_value is not None:
            self.value = new_value  # check that it is equal
        return self  # no need to create another instance for a constant

    def recombine(self: P, *others: P) -> None:
        pass

    def mutate(self) -> None:
        pass


def as_parameter(param: tp.Any) -> Parameter:
    """Returns a Parameter from anything:
    either the input if it is already a parameter, or a Constant if not
    This is convenient for iterating over Parameter and other objects alike
    """
    if isinstance(param, Parameter):
        return param
    else:
        return Constant(param)


class MultiobjectiveReference(Constant):
    def __init__(self, parameter: tp.Optional[Parameter] = None) -> None:
        if parameter is not None and not isinstance(parameter, Parameter):
            raise TypeError(
                "MultiobjectiveReference should either take no argument or a parameter which will "
                f"be used by the optimizer.\n(received {parameter} of type {type(parameter)})"
            )
        super().__init__(parameter)


class Operator(Layered):
    """Layer object that can be used as an operator on a Parameter"""

    _LAYER_LEVEL = Level.OPERATION

    def __call__(self, parameter: Parameter) -> Parameter:
        """Applies the operator on a Parameter to create a new Parameter"""
        new = parameter.copy()
        new.add_layer(self.copy())
        return new



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest
import numpy as np
import sys
from unittest import SkipTest
import nevergrad as ng
import nevergrad.common.typing as tp
from nevergrad.common import testing
from . import base
from .optimizerlib import registry


# decorators to be used when testing on Windows is unecessary
# or cumbersome
skip_win_perf = pytest.mark.skipif(
    sys.platform == "win32", reason="Slow, and no need to test performance on all platforms"
)


def suggestable(name: str) -> bool:
    # Some methods are not good with suggestions.
    keywords = ["TBPSA", "BO", "EMNA", "EDA", "BO", "Stupid", "Pymoo", "GOMEA"]
    return not any(x in name for x in keywords)


def suggestion_testing(
    name: str,
    instrumentation: ng.p.Array,
    suggestion: np.ndarray,
    budget: int,
    objective_function: tp.Callable[..., tp.Any],
    optimum: tp.Optional[np.ndarray] = None,
    threshold: tp.Optional[float] = None,
):
    optimizer_cls = registry[name]
    optim = optimizer_cls(instrumentation, budget)
    if optimum is None:
        optimum = suggestion
    optim.suggest(suggestion)
    optim.minimize(objective_function)
    if threshold is not None:
        assert (
            objective_function(optim.recommend().value) < threshold
        ), f"{name} proposes {optim.recommend().value} instead of {optimum} (threshold={threshold})"
        return
    assert np.all(
        optim.recommend().value == optimum
    ), f"{name} proposes {optim.recommend().value} instead of {optimum}"


@skip_win_perf  # type: ignore
@pytest.mark.parametrize("name", [r for r in registry if suggestable(r)])  # type: ignore
def test_suggest_optimizers(name: str) -> None:
    """Checks that each optimizer is able to converge when optimum is given"""

    if sum([ord(c) for c in name]) % 4 > 0 and name not in ["CMA", "PSO", "DE"]:
        raise SkipTest("Too expensive: we randomly skip 3/4 of these tests.")

    instrum = ng.p.Array(shape=(100,)).set_bounds(0.0, 1.0)
    instrum.set_integer_casting()
    suggestion = np.asarray([0] * 17 + [1] * 17 + [0] * 66)  # The optimum is the suggestion.
    target = lambda x: 0 if np.all(np.asarray(x, dtype=int) == suggestion) else 1
    suggestion_testing(name, instrum, suggestion, 7, target)


def good_at_suggest(name: str) -> bool:
    keywords = [
        "Noisy",
        "Optimistic",
        "DiscreteDE",
        "Multi",
        "Anisotropic",
        "BSO",
        "GOMEA",
        "Sparse",
        "Adaptive",
        "Doerr",
        "Recombining",
        "SA",
        "Lognormal",
        "PortfolioDiscreteOne",
        "FastGADiscreteOne",
    ]
    return not any(k in name for k in keywords)


@skip_win_perf  # type: ignore
@pytest.mark.parametrize("name", [r for r in registry if "iscre" in r and "Smooth" not in r and good_at_suggest(r) and r != "DiscreteOnePlusOne" and ("Lengler" not in r or "LenglerOne" in r)])  # type: ignore
def test_harder_suggest_optimizers(name: str) -> None:
    """Checks that discrete optimizers are good when a suggestion is nearby."""
    instrum = ng.p.Array(shape=(100,)).set_bounds(0.0, 1.0)
    instrum.set_integer_casting()
    optimum = np.asarray([0] * 17 + [1] * 17 + [0] * 66)
    target = lambda x: min(3, np.sum((np.asarray(x, dtype=int) - optimum) ** 2))
    suggestion = np.asarray([0] * 17 + [1] * 16 + [0] * 67)
    suggestion_testing(name, instrum, suggestion, 1500, target, optimum)


@skip_win_perf  # type: ignore
def test_harder_continuous_suggest_optimizers() -> None:
    """Checks that somes optimizer can converge when provided with a good suggestion."""
    instrum = ng.p.Array(shape=(100,)).set_bounds(0.0, 1.0)
    optimum = np.asarray([0] * 17 + [1] * 17 + [0] * 66)
    target = lambda x: min(2.0, np.sum((x - optimum) ** 2))
    suggestion = np.asarray([0] * 17 + [1] * 16 + [0] * 67)
    suggestion_testing("NGOpt", instrum, suggestion, 1500, target, optimum, threshold=0.9)


@testing.suppress_nevergrad_warnings()
@pytest.mark.parametrize("name", registry)  # type: ignore
def test_optimizers_suggest(name: str) -> None:  # pylint: disable=redefined-outer-name
    optimizer = registry[name](parametrization=4, budget=2)
    optimizer.suggest(np.array([12.0] * 4))
    candidate = optimizer.ask()
    try:
        optimizer.tell(candidate, 12)
        # The optimizer should recommend its suggestion, except for a few optimization methods:
        if name not in ["SPSA", "TBPSA", "StupidRandom"]:
            np.testing.assert_array_almost_equal(optimizer.provide_recommendation().value, [12.0] * 4)
    except base.errors.TellNotAskedNotSupportedError:
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import re
import tempfile
import operator
import contextlib
from pathlib import Path
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import testing
from . import utils


LINETOKEN = "@nevergrad" + "@"  # Do not trigger an error when parsing this file...
COMMENT_CHARS = {".c": "//", ".h": "//", ".cpp": "//", ".hpp": "//", ".py": "#", ".m": "%"}


def _convert_to_string(data: tp.Any, extension: str) -> str:
    """Converts the data into a string to be injected in a file"""
    if isinstance(data, np.ndarray):
        string = repr(data.tolist())
    else:
        string = repr(data)
    if extension in [".h", ".hpp", ".cpp", ".c"] and isinstance(
        data, np.ndarray
    ):  # TODO: custom extensions are handled as python
        string = string.replace("[", "{").replace("]", "}")
    return string


class Placeholder:
    """Placeholder tokens to for external code instrumentation"""

    pattern = r"NG_ARG" + r"{(?P<name>\w+?)(\|(?P<comment>.+?))?}"

    def __init__(self, name: str, comment: tp.Optional[str]) -> None:
        self.name = name
        self.comment = comment

    @classmethod
    def finditer(cls, text: str) -> tp.List["Placeholder"]:
        prog = re.compile(cls.pattern)
        return [cls(x.group("name"), x.group("comment")) for x in prog.finditer(text)]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name!a}, {self.comment!a})"

    def __eq__(self, other: tp.Any) -> bool:
        if self.__class__ == other.__class__:
            return (self.name, self.comment) == (other.name, other.comment)
        return False

    @classmethod
    def sub(cls, text: str, extension: str, replacers: tp.Dict[str, tp.Any]) -> str:
        found: tp.Set[str] = set()
        kwargs = {x: _convert_to_string(y, extension) for x, y in replacers.items()}

        def _replacer(regex: tp.Match[str]) -> str:
            name = regex.group("name")
            if name in found:
                raise RuntimeError(f'Trying to remplace a second time placeholder "{name}"')
            if name not in kwargs:
                raise KeyError(f'Could not find a value for placeholder "{name}"')
            found.add(name)
            return str(kwargs[name])

        text = re.sub(cls.pattern, _replacer, text)
        missing = set(kwargs) - found
        if missing:
            raise RuntimeError(f"All values have not been consumed: {missing}")
        return text


def symlink_folder_tree(folder: tp.Union[Path, str], shadow_folder: tp.Union[Path, str]) -> None:
    """Utility for copying the tree structure of a folder and symlinking all files
    This can help creating lightweight copies of a project, for instantiating several
    copies with different parameters.
    """
    folder, shadow_folder = (Path(x).expanduser().resolve().absolute() for x in (folder, shadow_folder))
    shadow_folder.mkdir(parents=True, exist_ok=True)
    for fp in folder.iterdir():  # iterating is more efficient than globbing here
        shadow_fp = shadow_folder / fp.name
        if fp.is_dir():
            symlink_folder_tree(fp, shadow_fp)
        elif not shadow_fp.exists():
            shadow_fp.symlink_to(fp)


def uncomment_line(line: str, extension: str) -> str:
    if extension not in COMMENT_CHARS:
        raise RuntimeError(
            f"Unknown file type: {extension}\nDid you register it using {FolderFunction.register_file_type.__name__}?"
        )
    pattern = r"^(?P<indent> *)"
    pattern += r"(?P<linetoken>" + COMMENT_CHARS[extension] + r" *" + LINETOKEN + r" *)"
    pattern += r"(?P<command>.*)"
    lineseg = re.search(pattern, line)
    if lineseg is not None:
        line = lineseg.group("indent") + lineseg.group("command")
    if LINETOKEN in line:
        raise RuntimeError(
            f"Uncommenting failed for line of {extension} file (a {LINETOKEN} tag remains):\n{line}\n"
            f"Did you follow the pattern indent+comment+{LINETOKEN}+code (with nothing before the indent)?"
        )
    return line


class FileTextFunction:
    """Function created from a file and generating the text file after
    replacement of the placeholders
    """

    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath
        assert filepath.exists(), "{filepath} does not exist"
        with filepath.open("r") as f:
            text = f.read()
        deprecated_placeholders = ["NG_G{", "NG_OD{", "NG_SC{"]
        if any(x in text for x in deprecated_placeholders):
            raise RuntimeError(
                f"Found one of deprecated placeholders {deprecated_placeholders}. The API has now evolved to "
                "a single placeholder NG_ARG{name|comment}, and FolderFunction now takes as many kwargs "
                "as placeholders and must be instrumented before optimization.\n"
                "Please refer to the README, PR #73 or issue #45 for more information"
            )
        if LINETOKEN in text:
            lines = text.splitlines()
            ext = filepath.suffix.lower()
            lines = [(l if LINETOKEN not in l else uncomment_line(l, ext)) for l in lines]
            text = "\n".join(lines)
        self.placeholders = Placeholder.finditer(text)
        self._text = text
        self.parameters: tp.Set[str] = set()
        for x in self.placeholders:
            if x.name not in self.parameters:
                self.parameters.add(x.name)
            else:
                raise RuntimeError(
                    f'Found duplicate placeholder (names must be unique) with name "{x.name}" in file:\n{self.filepath}'
                )

    def __call__(self, **kwargs: tp.Any) -> str:
        testing.assert_set_equal(kwargs, self.parameters, err_msg="Wrong input parameters.")
        return Placeholder.sub(self._text, self.filepath.suffix, replacers=kwargs)

    def __repr__(self) -> str:
        names = sorted(self.parameters)
        return f"{self.__class__.__name__}({self.filepath})({', '.join(names)})"


class FolderInstantiator:
    """Folder with instrumentation tokens, which can be instantiated.

    Parameters
    ----------
    folder: str/Path
        the instrumented folder to instantiate
    clean_copy: bool
        whether to create an initial clean temporary copy of the folder in order to avoid
        versioning problems (instantiations are lightweight symlinks in any case).

    Caution
    -------
        The clean copy is generally located in /tmp and may not be accessible for
        computation in a cluster. You may want to create a clean copy yourself
        in the folder of your choice, or set the the TemporaryDirectoryCopy class
        (located in instrumentation.instantiate) CLEAN_COPY_DIRECTORY environment
        variable to a shared directory
    """

    def __init__(self, folder: tp.Union[Path, str], clean_copy: bool = False) -> None:
        self._clean_copy = None
        self.folder = Path(folder).expanduser().absolute()
        assert self.folder.exists(), f"{folder} does not seem to exist"
        if clean_copy:
            self._clean_copy = utils.TemporaryDirectoryCopy(str(folder))
            self.folder = self._clean_copy.copyname
        self.file_functions: tp.List[FileTextFunction] = []
        names: tp.Set[str] = set()
        for fp in self.folder.glob("**/*"):  # TODO filter out all hidden files (+ build files?)
            if fp.is_file() and fp.suffix.lower() in COMMENT_CHARS:
                file_func = FileTextFunction(fp)
                fnames = {ph.name for ph in file_func.placeholders}
                if fnames:
                    if fnames & names:
                        raise RuntimeError(
                            f"Found {fp} placeholders in another file (names must be unique): {fnames & names}"
                        )
                    self.file_functions.append(file_func)
        assert self.file_functions, "Found no file with placeholders"
        self.file_functions = sorted(self.file_functions, key=operator.attrgetter("filepath"))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.folder}") with files:\n{self.file_functions}'

    @property
    def placeholders(self) -> tp.List[Placeholder]:
        return [p for f in self.file_functions for p in f.placeholders]

    def instantiate_to_folder(self, outfolder: tp.Union[Path, str], kwargs: tp.Dict[str, tp.Any]) -> None:
        testing.assert_set_equal(
            kwargs, {x.name for x in self.placeholders}, err_msg="Wrong input parameters."
        )
        outfolder = Path(outfolder).expanduser().absolute()
        assert outfolder != self.folder, "Do not instantiate on same folder!"
        symlink_folder_tree(self.folder, outfolder)
        for file_func in self.file_functions:
            inst_fp = outfolder / file_func.filepath.relative_to(self.folder)
            os.remove(str(inst_fp))  # remove symlink to avoid writing in original dir
            with inst_fp.open("w") as f:
                f.write(file_func(**{x: y for x, y in kwargs.items() if x in file_func.parameters}))

    @contextlib.contextmanager
    def instantiate(self, **kwargs: tp.Any) -> tp.Generator[Path, None, None]:
        with tempfile.TemporaryDirectory() as tempfolder:
            subtempfolder = Path(tempfolder) / self.folder.name
            self.instantiate_to_folder(subtempfolder, kwargs)
            yield subtempfolder


class FolderFunction:
    """Turns a folder into a parametrized function
    (with nevergrad tokens)

    Parameters
    ----------
    folder: Path/str
        path to the folder to instrument
    command: list
        command to run from inside the folder. The last line in stdout will
        be the output of the function.
        The command must be performed from just outside the instrument
        directory
    verbose: bool
        whether to print the run command and from where it is run.
    clean_copy: bool
        whether to create an initial clean temporary copy of the folder in order to avoid
        versioning problems (instantiations are lightweight symlinks in any case)

    Returns
    -------
    Any
        the post-processed output of the called command

    Note
    ----
    By default, the postprocessing attribute holds a function which recovers the last line
    and converts it to float. The sequence of postprocessing can however be tampered
    with directly in order to change it

    Caution
    -------
        The clean copy is generally located in /tmp and may not be accessible for
        computation in a cluster. You may want to create a clean copy yourself
        in the folder of your choice, or set the the TemporaryDirectoryCopy class
        (located in instrumentation.instantiate) CLEAN_COPY_DIRECTORY environment
        variable to a shared directory
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        folder: tp.Union[Path, str],
        command: tp.List[str],
        verbose: bool = False,
        clean_copy: bool = False,
    ) -> None:
        self.command = command
        self.verbose = verbose
        self.postprocessings = [get_last_line_as_float]
        self.instantiator = FolderInstantiator(folder, clean_copy=clean_copy)
        self.last_full_output: tp.Optional[str] = None

    @staticmethod
    def register_file_type(suffix: str, comment_chars: str) -> None:
        """Register a new file type to be used for token instrumentation by providing the relevant file suffix as well as
        the characters that indicate a comment."""
        if not suffix.startswith("."):
            suffix = f".{suffix}"
        COMMENT_CHARS[suffix] = comment_chars

    @property
    def placeholders(self) -> tp.List[Placeholder]:
        return self.instantiator.placeholders

    def __call__(self, **kwargs: tp.Any) -> tp.Any:
        with self.instantiator.instantiate(**kwargs) as folder:
            if self.verbose:
                print(f"Running {self.command} from {folder.parent} which holds {folder}")
            output: tp.Any = utils.CommandFunction(self.command, cwd=folder.parent)()
        if self.verbose:
            print(f"FolderFunction recovered full output:\n{output}")
        self.last_full_output = output.strip()
        if not output:
            raise ValueError("No output")
        for postproc in self.postprocessings:
            output = postproc(output)
        if self.verbose:
            print(f"FolderFunction returns: {output}")
        return output


def get_last_line_as_float(output: str) -> float:
    split_output = output.strip().splitlines()
    return float(split_output[-1])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from functools import partial
import numpy as np
import pyomo.environ as pyomo
import nevergrad.common.typing as tp
from nevergrad.parametrization import parameter as p
from .. import base


ParamDict = tp.Dict[str, p.Parameter]


def _convert_to_ng_name(pyomo_var_key: tp.Any) -> str:
    if isinstance(pyomo_var_key, str):
        return '"' + str(pyomo_var_key) + '"'
    else:
        return str(pyomo_var_key)


def _make_pyomo_range_set_to_parametrization(
    domain: pyomo.RangeSet, params: ParamDict, params_name: str
) -> ParamDict:
    # https://pyomo.readthedocs.io/en/stable/pyomo_modeling_components/Sets.html
    # Refer to the implementation in pyomo/core/base/set.py
    ranges = list(domain.ranges())
    num_ranges = len(ranges)
    if num_ranges == 1 and (ranges[0].step in [-1, 0, 1]):
        if isinstance(ranges[0], pyomo.base.range.NumericRange):
            lb, ub = ranges[0].start, ranges[0].end
            if ranges[0].step < 0:
                lb, ub = ub, lb
            if (lb is not None) and (not ranges[0].closed[0]):
                lb = float(np.nextafter(lb, 1))
            if (ub is not None) and (not ranges[0].closed[1]):
                ub = float(np.nextafter(ub, -1))
            params[params_name] = p.Scalar(lower=lb, upper=ub)
            if ranges[0].step in [-1, 1]:
                # May consider using nested param
                params[params_name].set_integer_casting()  # type: ignore
        else:
            raise NotImplementedError(f"Cannot handle range type {type(ranges[0])}")
    elif isinstance(domain, pyomo.FiniteSimpleRangeSet):
        # Need to handle step size
        params[params_name] = p.Choice(
            [range(*r) for r in domain.ranges()]
        )  # Assume the ranges do not overlapped
    else:
        raise NotImplementedError(f"Cannot handle domain type {type(domain)}")
    return params


def _make_pyomo_variable_to_parametrization(model_component: pyomo.Var, params: ParamDict) -> ParamDict:
    # https://pyomo.readthedocs.io/en/stable/pyomo_modeling_components/Sets.html
    # Refer to the implementation in pyomo/core/base/var.py
    # To further improve the readability function, we should find out how to represent {None: ng.p.Scalar(), 1: ng.p.Scalar()} in ng.p.Dict
    # We do not adopt nested parameterization, which will require type information between string and int.
    # Such conversion has to be done in _pyomo_obj_function_wrapper and _pyomo_constraint_wrapper, which slows down optimization.
    if not isinstance(model_component, (pyomo.base.var.IndexedVar, pyomo.base.var.SimpleVar)):
        raise NotImplementedError  # Normally, Pyomo will create a set for the indices used by a variable
    for k, v in model_component._data.items():
        if isinstance(v, pyomo.base.var._GeneralVarData):
            if v.is_fixed():
                raise NotImplementedError
            if k is None:
                params_name = str(model_component.name)
            else:
                params_name = f"{model_component.name}[{_convert_to_ng_name(k)}]"
            if isinstance(v.domain, pyomo.RangeSet):
                params = _make_pyomo_range_set_to_parametrization(v.domain, params, params_name)
            elif isinstance(v.domain, pyomo.Set) and v.domain.isfinite():
                if v.domain.isordered():
                    params[params_name] = p.Choice(list(v.domain.ordered_data()))
                else:
                    params[params_name] = p.Choice(list(v.domain.data()))
            else:
                raise NotImplementedError(f"Cannot handle domain type {type(v.domain)}")
        else:
            raise NotImplementedError(f"Cannot handle variable type {type(v)}")
    return params


class Pyomo(base.ExperimentFunction):
    """Function calling Pyomo model

    Parameters
    ----------
    model: pyomo.environ.model
        Pyomo model

    Returns
    -------
    float
        the fitness

    Notes
    -----
    - You will require an Pyomo installation (with pip: "pip install pyomo")
    - Any changes on the model externally can lead to unexpected behaviours.
    """

    def __init__(self, model: pyomo.Model) -> None:
        if isinstance(model, pyomo.ConcreteModel):
            self._model_instance = model.clone()  # To enable the objective function to run in parallel
        else:
            raise NotImplementedError(
                "AbstractModel is not supported. Please use create_instance() in Pyomo to create a model instance."
            )

        instru_params: ParamDict = {}
        self.all_vars: tp.List[pyomo.Var] = []
        self.all_params: tp.List[pyomo.Param] = []
        self.all_constraints: tp.List[pyomo.Constraint] = []
        self.all_objectives: tp.List[pyomo.Objective] = []

        # Relevant document: https://pyomo.readthedocs.io/en/stable/working_models.html

        for v in self._model_instance.component_objects(pyomo.Var, active=True):
            self.all_vars.append(v)
            _make_pyomo_variable_to_parametrization(v, instru_params)
        for v in self._model_instance.component_objects(pyomo.Param, active=True):
            self.all_params.append(v)
        for v in self._model_instance.component_objects(pyomo.Constraint, active=True):
            self.all_constraints.append(v)
        for v in self._model_instance.component_objects(pyomo.Objective, active=True):
            # if v.sense == -1:
            #    print(f"Only minimization problem is supported. The value of the objective function {v.name} will be multiplied by -1.")
            self.all_objectives.append(v)

        if not self.all_objectives:
            raise NotImplementedError("Cannot find objective function")

        if len(self.all_objectives) > 1:
            raise NotImplementedError("Multi-objective function is not supported yet.")

        self._value_assignment_code_obj = ""

        instru = p.Instrumentation(**instru_params).set_name("")
        for c_idx in range(0, len(self.all_constraints)):
            instru.register_cheap_constraint(partial(self._pyomo_constraint_wrapper, c_idx))
        super().__init__(
            function=partial(self._pyomo_obj_function_wrapper, 0), parametrization=instru
        )  # Single objective

        exp_tag = ",".join([n.name for n in self.all_objectives])
        exp_tag += "|" + ",".join([n.name for n in self.all_vars])
        exp_tag += "|" + ",".join([n.name for n in self.all_constraints])
        self.add_descriptors(name=exp_tag)

    def _pyomo_value_assignment(self, k_model_variables: tp.Dict[str, tp.Any]) -> None:
        if self._value_assignment_code_obj == "":
            code_str = ""
            for k in k_model_variables:
                code_str += f"self._model_instance.{k} = k_model_variables['{k}']\n"
            self._value_assignment_code_obj = compile(code_str, "<string>", "exec")  # type: ignore
        # TODO find a way to avoid exec
        exec(self._value_assignment_code_obj)  # pylint: disable=exec-used

    def _pyomo_obj_function_wrapper(self, i: int, **k_model_variables: tp.Dict[str, tp.Any]) -> float:
        self._pyomo_value_assignment(k_model_variables)
        return float(
            pyomo.value(self.all_objectives[i] * self.all_objectives[i].sense)
        )  # Single objective assumption

    def _pyomo_constraint_wrapper(self, i: int, instru: tp.ArgsKwargs) -> bool:
        k_model_variables = instru[1]
        # Combine all constraints into single one
        self._pyomo_value_assignment(k_model_variables)
        if isinstance(self.all_constraints[i], pyomo.base.constraint.SimpleConstraint):
            return bool(pyomo.value(self.all_constraints[i].expr(self._model_instance)))
        elif isinstance(self.all_constraints[i], pyomo.base.constraint.IndexedConstraint):
            ret = True
            for _, c in self.all_constraints[i].items():
                ret = ret and pyomo.value(c.expr(self._model_instance))
                if not ret:
                    break
            return ret
        else:
            raise NotImplementedError(
                f"Constraint type {self.all_constraints[i].ctype} is not supported yet."
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.(an
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# import warnings
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import errors

from . import _layering
from . import core
from .container import Dict
from . import utils


# pylint: disable=no-value-for-parameter,import-outside-toplevel
# pylint: disable=cyclic-import


D = tp.TypeVar("D", bound="Data")
P = tp.TypeVar("P", bound=core.Parameter)


def _param_string(parameters: Dict) -> str:
    """Hacky helper for nice-visualizatioon"""
    substr = f"[{parameters._get_parameters_str()}]"
    if substr == "[]":
        substr = ""
    return substr


# pylint: disable=too-many-arguments, too-many-instance-attributes,abstract-method
class Data(core.Parameter):
    """Array/scalar parameter

    Parameters
    ----------
    init: np.ndarray, or None
        initial value of the array (defaults to 0, with a provided shape)
    shape: tuple of ints, or None
        shape of the array, to be provided iff init is not provided
    lower: array, float or None
        minimum value
    upper: array, float or None
        maximum value
    mutable_sigma: bool
        whether the mutation standard deviation must mutate as well (for mutation based algorithms)

    Note
    ----
    - More specific behaviors can be obtained throught the following methods:
      set_bounds, set_mutation, set_integer_casting
    - if both lower and upper bounds are provided, sigma will be adapted so that the range spans 6 sigma.
      Also, if init is not provided, it will be set to the middle value.
    """

    def __init__(
        self,
        *,
        init: tp.Optional[tp.ArrayLike] = None,
        shape: tp.Optional[tp.Tuple[int, ...]] = None,
        lower: tp.BoundValue = None,
        upper: tp.BoundValue = None,
        mutable_sigma: bool = False,
    ) -> None:
        super().__init__()
        sigma: tp.Any = np.array([1.0])
        # make sure either shape or
        if isinstance(lower, (list, tuple)):
            lower = np.array(lower, dtype=float)
        if isinstance(upper, (list, tuple)):
            upper = np.array(upper, dtype=float)
        if sum(x is None for x in [init, shape]) != 1:
            raise ValueError('Exactly one of "init" or "shape" must be provided')
        if init is not None:
            init = np.array(init, dtype=float, copy=False)
        else:
            assert isinstance(shape, (list, tuple)) and all(
                isinstance(n, int) for n in shape
            ), f"Incorrect shape: {shape} (type: {type(shape)})."
            init = np.zeros(shape, dtype=float)
            if lower is not None and upper is not None:
                init += (lower + upper) / 2.0
        self._value = init
        self.add_layer(_layering.ArrayCasting())
        # handle bounds
        num_bounds = sum(x is not None for x in (lower, upper))
        layer: tp.Any = None
        if num_bounds:
            from . import _datalayers

            layer = _datalayers.Bound(
                lower=lower, upper=upper, uniform_sampling=init is None and num_bounds == 2
            )
            if num_bounds == 2:
                sigma = (layer.bounds[1] - layer.bounds[0]) / 6
        # set parameters
        sigma = sigma[0] if sigma.size == 1 else sigma
        if mutable_sigma:
            siginit = sigma
            # for the choice of the base:
            # cf Evolution Strategies, Hans-Georg Beyer (2007)
            # http://www.scholarpedia.org/article/Evolution_strategies
            # we want:
            # sigma *= exp(gaussian / sqrt(dim))
            base = float(np.exp(1.0 / np.sqrt(2 * init.size)))
            sigma = base ** (Array if isinstance(sigma, np.ndarray) else Scalar)(
                init=siginit, mutable_sigma=False
            )
            sigma.value = siginit
        self.parameters = Dict(sigma=sigma)
        self.parameters._ignore_in_repr = dict(sigma="1.0")
        if layer is not None:
            layer(self, inplace=True)

    @property
    def bounds(self) -> tp.Tuple[tp.Optional[np.ndarray], tp.Optional[np.ndarray]]:
        """Estimate of the bounds (None if unbounded)

        Note
        ----
        This may be inaccurate (WIP)
        """
        from . import _datalayers

        bound_layers = _datalayers.BoundLayer.filter_from(self)
        if not bound_layers:
            return (None, None)
        bounds = bound_layers[-1].bounds
        forwardable = _datalayers.ForwardableOperation.filter_from(self)
        forwardable = [x for x in forwardable if x._layer_index > bound_layers[-1]._layer_index]
        for f in forwardable:
            bounds = tuple(None if b is None else f.forward(b) for b in bounds)  # type: ignore
        return bounds

    @property
    def dimension(self) -> int:
        return int(np.prod(self._value.shape))

    def _get_name(self) -> str:
        cls = self.__class__.__name__
        descriptors: tp.List[str] = []
        if self._value.shape != (1,):
            descriptors.append(str(self._value.shape).replace(" ", ""))
        descriptors += [
            layer.name
            for layer in self._layers[1:]
            if not isinstance(layer, (_layering.ArrayCasting, _layering._ScalarCasting))
        ]
        description = "" if not descriptors else "{{{}}}".format(",".join(descriptors))
        return f"{cls}{description}" + _param_string(self.parameters)

    @property
    def sigma(self) -> "Data":
        """Value for the standard deviation used to mutate the parameter"""
        return self.parameters["sigma"]  # type: ignore

    def _layered_sample(self: D) -> D:
        child = self.spawn_child()
        from . import helpers

        with helpers.deterministic_sampling(child):
            child.mutate()
        return child

    # pylint: disable=unused-argument
    def set_bounds(
        self: D,
        lower: tp.BoundValue = None,
        upper: tp.BoundValue = None,
        method: str = "bouncing",
        full_range_sampling: tp.Optional[bool] = None,
    ) -> D:
        """Bounds all real values into [lower, upper] using a provided method

        Parameters
        ----------
        lower: array, float or None
            minimum value
        upper: array, float or None
            maximum value
        method: str
            One of the following choices:

            - "bouncing": bounce on border (at most once). This is a variant of clipping,
               avoiding bounds over-samping (default).
            - "clipping": clips the values inside the bounds. This is efficient but leads
              to over-sampling on the bounds.
            - "constraint": adds a constraint (see register_cheap_constraint) which leads to rejecting mutations
              reaching beyond the bounds. This avoids oversampling the boundaries, but can be inefficient in large
              dimension.
            - "arctan": maps the space [lower, upper] to to all [-inf, inf] using arctan transform. This is efficient
              but it completely reshapes the space (a mutation in the center of the space will be larger than a mutation
              close to the bounds), and reaching the bounds is equivalent to reaching the infinity.
            - "tanh": same as "arctan", but with a "tanh" transform. "tanh" saturating much faster than "arctan", it can lead
              to unexpected behaviors.
        full_range_sampling: Optional bool
            Changes the default behavior of the "sample" method (aka creating a child and mutating it from the current instance)
            or the sampling optimizers, to creating a child with a value sampled uniformly (or log-uniformly) within
            the while range of the bounds. The "sample" method is used by some algorithms to create an initial population.
            This is activated by default if both bounds are provided.

        Notes
        -----
        - "tanh" reaches the boundaries really quickly, while "arctan" is much softer
        - only "clipping" accepts partial bounds (None values)
        """  # TODO improve description of methods
        from . import _datalayers

        # if method == "constraint":
        #     method = "clipping"
        value = self.value
        if method == "constraint":
            layer = _datalayers.BoundLayer(lower=lower, upper=upper, uniform_sampling=full_range_sampling)
            checker = utils.BoundChecker(*layer.bounds)
            self.register_cheap_constraint(checker)
        else:
            layer = _datalayers.Bound(
                lower=lower, upper=upper, method=method, uniform_sampling=full_range_sampling
            )
        layer._LEGACY = True
        layer(self, inplace=True)
        _fix_legacy(self)
        try:
            self.value = value
        except ValueError as e:
            raise errors.NevergradValueError(
                "Current value is not within bounds, please update it first"
            ) from e
        return self

    def set_mutation(
        self: D,
        sigma: tp.Optional[tp.Union[float, core.Parameter]] = None,
        exponent: tp.Optional[float] = None,
    ) -> D:
        """Output will be cast to integer(s) through deterministic rounding.

        Parameters
        ----------
        sigma: Array/Log or float
            The standard deviation of the mutation. If a Parameter is provided, it will replace the current
            value. If a float is provided, it will either replace a previous float value, or update the value
            of the Parameter.
        exponent: float
            exponent for the logarithmic mode. With the default sigma=1, using exponent=2 will perform
            x2 or /2 "on average" on the value at each mutation.

        Returns
        -------
        self
        """
        if sigma is not None:
            # just replace if an actual Parameter is provided as sigma, else update value (parametrized or not)
            if isinstance(sigma, core.Parameter) or isinstance(
                self.parameters._content["sigma"], core.Constant
            ):
                self.parameters._content["sigma"] = core.as_parameter(sigma)
            else:
                self.sigma.value = sigma
        if exponent is not None:
            from . import _datalayers

            if exponent <= 0.0:
                raise ValueError("Only exponents strictly higher than 0.0 are allowed")
            value = self.value
            layer = _datalayers.Exponent(base=exponent)
            layer._LEGACY = True
            self.add_layer(layer)
            _fix_legacy(self)
            try:
                self.value = value
            except ValueError as e:
                raise errors.NevergradValueError(
                    "Cannot convert to logarithmic mode with current non-positive value, please update it firstp."
                ) from e
        return self

    def set_integer_casting(self: D) -> D:
        """Output will be cast to integer(s) through deterministic rounding.

        Returns
        -------
        self

        Note
        ----
        Using integer casting makes the parameter discrete which can make the optimization more
        difficult. It is especially ill-advised to use this with a range smaller than 10, or
        a sigma lower than 1. In those cases, you should rather use a TransitionChoice instead.
        """
        return self.add_layer(_layering.Int())

    @property
    def integer(self) -> bool:
        return any(isinstance(x, _layering.Int) for x in self._layers)

    # pylint: disable=unused-argument
    def _internal_set_standardized_data(self: D, data: np.ndarray, reference: D) -> None:
        assert isinstance(data, np.ndarray)
        sigma = reference.sigma.value
        data_reduc = sigma * (data + reference._to_reduced_space()).reshape(reference._value.shape)
        self._value = data_reduc
        # make sure _value is updated by the layers getters if need be:
        self.value  # pylint: disable=pointless-statement

    def _internal_get_standardized_data(self: D, reference: D) -> np.ndarray:
        return reference._to_reduced_space(self._value - reference._value)

    def _to_reduced_space(self, value: tp.Optional[np.ndarray] = None) -> np.ndarray:
        """Converts array with appropriate shapes to reduced (uncentered) space
        by applying log scaling and sigma scaling
        """
        # TODO this is nearly useless now that the layer system has been added. Remove?
        if value is None:
            value = self._value
        reduced = value / self.sigma.value
        return reduced.ravel()  # type: ignore

    def _layered_recombine(self: D, *others: D) -> None:  # type: ignore
        all_params = [self] + list(others)
        all_arrays = [p.get_standardized_data(reference=self) for p in all_params]
        mean: np.ndarray = np.mean(all_arrays, axis=0)  # type: ignore
        self.set_standardized_data(mean)

    def copy(self: D) -> D:
        child = super().copy()
        child._value = np.array(self._value, copy=True)
        return child

    def _layered_set_value(self, value: np.ndarray) -> None:
        self._check_frozen()
        if self._value.shape != value.shape:
            raise ValueError(
                f"Cannot set array of shape {self._value.shape} with value of shape {value.shape}"
            )
        self._value = value

    def _layered_get_value(self) -> np.ndarray:
        return self._value

    def _new_with_data_layer(self: D, name: str, *args: tp.Any, **kwargs: tp.Any) -> D:
        # pylint: disable=cyclic-import
        from . import _datalayers  # lazy to avoid cyclic imports

        new = self.copy()
        new.add_layer(getattr(_datalayers, name)(*args, **kwargs))
        return new

    def __mod__(self: D, module: tp.Any) -> D:
        return self._new_with_data_layer("Modulo", module)

    def __rpow__(self: D, base: float) -> D:
        return self._new_with_data_layer("Exponent", base)

    def __add__(self: D, offset: tp.Any) -> D:
        return self._new_with_data_layer("Add", offset)

    def __sub__(self: D, offset: tp.Any) -> D:
        return self.__add__(-offset)

    def __radd__(self: D, offset: tp.Any) -> D:
        return self.__add__(offset)

    def __mul__(self: D, value: tp.Any) -> D:
        return self._new_with_data_layer("Multiply", value)

    def __rmul__(self: D, value: tp.Any) -> D:
        return self.__mul__(value)

    def __truediv__(self: D, value: tp.Any) -> D:
        return self.__mul__(1.0 / value)

    def __rtruediv__(self: D, value: tp.Any) -> D:
        return value * (self**-1)  # type: ignore

    def __pow__(self: D, power: float) -> D:
        return self._new_with_data_layer("Power", power)

    def __neg__(self: D) -> D:
        return self.__mul__(-1.0)


def _fix_legacy(parameter: Data) -> None:
    """Ugly hack for keeping the legacy behaviors with considers bounds always after the exponent
    and can still sample "before" the exponent (log-uniform).
    """
    from . import _datalayers

    legacy = [x for x in _datalayers.Operation.filter_from(parameter) if x._LEGACY]
    if len(legacy) < 2:
        return
    if len(legacy) > 2:
        raise errors.NevergradRuntimeError("More than 2 legacy layers, this should not happen, open an issue")
    # warnings.warn(
    #     "Settings bounds and exponent through the Array/Scalar API will change behavior "
    #     " (this is an early warning, more on this asap)",
    #     errors.NevergradBehaviorChangesWarning,
    # )  # TODO activate when ready
    value = parameter.value
    layers_inds = tuple(leg._layer_index for leg in legacy)
    if abs(layers_inds[0] - layers_inds[1]) > 1:
        raise errors.NevergradRuntimeError("Non-legacy layers between 2 legacy layers")
    parameter._layers = [x for x in parameter._layers if x._layer_index not in layers_inds]
    # fix parameter layers
    for k, sub in enumerate(parameter._layers):
        sub._layer_index = k
        sub._layers = parameter._layers
    parameter.value = value
    bound_ind = int(isinstance(legacy[0], _datalayers.Exponent))
    bound: _datalayers.BoundLayer = legacy[bound_ind]  # type: ignore
    exp: _datalayers.Exponent = legacy[(bound_ind + 1) % 2]  # type: ignore
    bound.bounds = tuple(None if b is None else exp.backward(b) for b in bound.bounds)  # type: ignore
    if isinstance(bound, _datalayers.Bound):
        bound = _datalayers.Bound(
            lower=bound.bounds[0],
            upper=bound.bounds[1],
            method=bound._method,
            uniform_sampling=bound.uniform_sampling,
        )
    for lay in (bound, exp):
        lay._layer_index = 0
        lay._layers = [lay]
        parameter.add_layer(lay)
    return


class Array(Data):

    value: core.ValueProperty[tp.ArrayLike, np.ndarray] = core.ValueProperty()


class Scalar(Data):
    """Parameter representing a scalar.

    Parameters
    ----------
    init: optional float
        initial value of the scalar (defaults to 0.0 if both bounds are not provided)
    lower: optional float
        minimum value if any
    upper: optional float
        maximum value if any
    mutable_sigma: bool
        whether the mutation standard deviation must mutate as well (for mutation based algorithms)

    Notes
    -----
    - by default, this is an unbounded scalar with Gaussian mutations.
    - if both lower and upper bounds are provided, sigma will be adapted so that the range spans 6 sigma.
      Also, if init is not provided, it will be set to the middle value.
    - More specific behaviors can be obtained throught the following methods:
      :code:`set_bounds`, :code:`set_mutation`, :code:`set_integer_casting`
    """

    value: core.ValueProperty[float, float] = core.ValueProperty()

    def __init__(
        self,
        init: tp.Optional[float] = None,
        *,
        lower: tp.Optional[float] = None,
        upper: tp.Optional[float] = None,
        mutable_sigma: bool = True,
    ) -> None:
        bounded = all(a is not None for a in (lower, upper))
        no_init = init is None
        if bounded:
            if init is None:
                init = (lower + upper) / 2.0  # type: ignore
        if init is None:
            init = 0.0
        super().__init__(init=np.array([init]), mutable_sigma=mutable_sigma)
        if bounded:
            self.set_mutation(sigma=(upper - lower) / 6)  # type: ignore
        if any(a is not None for a in (lower, upper)):
            self.set_bounds(lower=lower, upper=upper, full_range_sampling=bounded and no_init)
        self.add_layer(_layering._ScalarCasting())


class Log(Scalar):
    """Parameter representing a positive variable, mutated by Gaussian mutation in log-scale.

    Parameters
    ----------
    init: float or None
        initial value of the variable. If not provided, it is set to the middle of lower and upper in log space
    exponent: float or None
        exponent for the log mutation: an exponent of 2.0 will lead to mutations by factors between around 0.5 and 2.0
        By default, it is set to either 2.0, or if the parameter is completely bounded to a factor so that bounds are
        at 3 sigma in the transformed space.
    lower: float or None
        minimum value if any (> 0)
    upper: float or None
        maximum value if any
    mutable_sigma: bool
        whether the mutation standard deviation must mutate as well (for mutation based algorithms)

    Note
    ----
    This class is only a wrapper over :code:`Scalar`.
    """

    def __init__(
        self,
        *,
        init: tp.Optional[float] = None,
        exponent: tp.Optional[float] = None,
        lower: tp.Optional[float] = None,
        upper: tp.Optional[float] = None,
        mutable_sigma: bool = False,
    ) -> None:
        no_init = init is None
        bounded = all(a is not None for a in (lower, upper))
        if bounded:
            if init is None:
                init = float(np.sqrt(lower * upper))  # type: ignore
            if exponent is None:
                exponent = float(
                    np.exp((np.log(upper) - np.log(lower)) / 6.0)  # type: ignore
                )  # 99.7% of values within the bounds
        if init is None:
            raise ValueError("You must define either a init value or both lower and upper bounds")
        if exponent is None:
            exponent = 2.0
        from . import _datalayers

        exp_layer = _datalayers.Exponent(exponent)
        raw_bounds = tuple(None if x is None else np.array([x], dtype=float) for x in (lower, upper))
        bounds = tuple(None if x is None else exp_layer.backward(x) for x in raw_bounds)
        init = exp_layer.backward(np.array([init]))
        super().__init__(init=init[0], mutable_sigma=mutable_sigma)  # type: ignore
        # TODO remove the next line when all compatibility is done
        if any(x is not None for x in bounds):
            bound_layer = _datalayers.Bound(*bounds, uniform_sampling=bounded and no_init)  # type: ignore
            bound_layer(self, inplace=True)
        self.add_layer(exp_layer)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import pytest
import numpy as np
import nevergrad as ng
from . import constraints


def function(*args: float) -> tp.Any:
    if len(args) == 1:
        return args[0]
    return args


@pytest.mark.parametrize("num", (1, 3))  # type: ignore
def test_constraint(num: int) -> None:
    parameter = ng.p.Instrumentation(*(ng.p.Scalar(np.random.randn()) for _ in range(num)))
    constrained = constraints.Constraint(function)(parameter)
    # check basic layer functionalities
    layer: constraints.Constraint = constrained._layers[-1]  # type: ignore
    assert layer.function(*([1] * num)) == 1 if num == 1 else [1] * num
    assert layer.function(*([-1] * num)) == 0 if num == 1 else [0] * num
    assert (
        np.mean([x < 0.1 for x in constrained.args]) > 0.5
    ), constrained.args  # some slack to avoid flakiness



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import typing as tp
import torch
import numpy as np


import cv2
from nevergrad.functions.base import UnsupportedExperiment as UnsupportedExperiment
from nevergrad.common.decorators import Registry


registry: Registry[tp.Any] = Registry()
MODELS: tp.Dict[str, tp.Any] = {}


class ImageLoss:

    REQUIRES_REFERENCE = True

    def __init__(self, reference: tp.Optional[np.ndarray] = None) -> None:
        if reference is not None:
            self.reference = reference
            assert len(self.reference.shape) == 3, self.reference.shape
            assert self.reference.min() >= 0.0
            assert self.reference.max() <= 256.0, f"Image max = {self.reference.max()}"
            assert self.reference.max() > 3.0  # Not totally sure but entirely black images are not very cool.
            self.domain_shape = self.reference.shape

    def __call__(self, img: np.ndarray) -> float:
        raise NotImplementedError(f"__call__ undefined in class {type(self)}")


@registry.register
class SumAbsoluteDifferences(ImageLoss):
    def __call__(self, x: np.ndarray) -> float:
        assert x.shape == self.domain_shape, f"Shape = {x.shape} vs {self.domain_shape}"
        value = float(np.sum(np.fabs(x - self.reference)))
        return value


class Lpips(ImageLoss):
    def __init__(self, reference: tp.Optional[np.ndarray] = None, net: str = "") -> None:
        super().__init__(reference)
        self.net = net
        # pylint: disable=import-outside-toplevel
        try:
            import lpips
        except ImportError:
            raise UnsupportedExperiment("LPIPS is not installed, please run 'pip install lpips'")
        self._LPIPS = lpips.LPIPS

    def __call__(self, img: np.ndarray) -> float:
        if self.net not in MODELS:
            MODELS[self.net] = self._LPIPS(net=self.net)
        loss_fn = MODELS[self.net]
        assert img.shape[2] == 3
        assert len(img.shape) == 3
        assert img.max() <= 256.0, f"Image max = {img.max()}"
        assert img.min() >= 0.0
        assert img.max() > 3.0
        img0 = torch.clamp(torch.Tensor(img).unsqueeze(0).permute(0, 3, 1, 2) / 256.0, 0, 1) * 2.0 - 1.0
        img1 = (
            torch.clamp(torch.Tensor(self.reference.copy()).unsqueeze(0).permute(0, 3, 1, 2) / 256.0, 0, 1)
            * 2.0
            - 1.0
        )  # The copy operation is here because of a warning otherwise, as Torch does not support non-writable numpy arrays.
        return float(loss_fn(img0, img1))


@registry.register
class LpipsAlex(Lpips):
    def __init__(self, reference: np.ndarray) -> None:
        super().__init__(reference, net="alex")


@registry.register
class LpipsVgg(Lpips):
    def __init__(self, reference: np.ndarray) -> None:
        super().__init__(reference, net="vgg")


@registry.register
class SumSquareDifferences(ImageLoss):
    def __call__(self, x: np.ndarray) -> float:
        assert x.shape == self.domain_shape, f"Shape = {x.shape} vs {self.domain_shape}"
        value = float(np.sum((x - self.reference) ** 2))
        return value


@registry.register
class HistogramDifference(ImageLoss):
    def __call__(self, x: np.ndarray) -> float:
        assert x.shape == self.domain_shape, f"Shape = {x.shape} vs {self.domain_shape}"
        assert x.shape[2] == 3
        x_gray_1d = np.sum(x, 2).ravel()
        ref_gray_1d = np.sum(self.reference, 2).ravel()
        value = float(np.sum(np.sort(x_gray_1d) - np.sort(ref_gray_1d)))
        return value


@registry.register
class Koncept512(ImageLoss):
    """
    This loss uses the neural network Koncept512 to score images
    It takes one image or a list of images of shape [x, y, 3], with each pixel between 0 and 256, and returns a score.
    """

    REQUIRES_REFERENCE = False

    @property
    def koncept(self) -> tp.Any:  # cache the model
        key = "koncept"
        if os.name == "nt":
            raise UnsupportedExperiment("Koncept512 is not working properly under Windows")
        if key not in MODELS:
            # pylint: disable=import-outside-toplevel
            try:
                from koncept.models import Koncept512 as K512Model
            except ImportError:
                raise UnsupportedExperiment("Koncept512 is not installed, please run 'pip install koncept'")
            MODELS[key] = K512Model()
        return MODELS[key]

    def __call__(self, img: np.ndarray) -> float:
        loss = -self.koncept.assess(img)
        return float(loss)


@registry.register
class Blur(ImageLoss):
    """
    This estimates bluriness.
    """

    REQUIRES_REFERENCE = False

    def __call__(self, img: np.ndarray) -> float:
        assert img.shape[2] == 3
        assert len(img.shape) == 3
        img = np.asarray(img, dtype=np.float64)
        return -float(cv2.Laplacian(img, cv2.CV_64F).var())  # type: ignore


@registry.register
class Brisque(ImageLoss):
    """
    This estimates the Brisque score (lower is better).
    """

    REQUIRES_REFERENCE = False

    def __call__(self, img: np.ndarray) -> float:
        try:
            import imquality.brisque as brisque
        except ImportError:
            raise UnsupportedExperiment("Brisque is not installed, please run 'pip install imquality'")
        try:
            score = brisque.score(img)
        except AssertionError:  # oh my god, brisque can raise an assert when the data is too weird.
            score = float("inf")
        return score



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
from unittest import TestCase
import numpy as np
from . import decorators


class DecoratorTests(TestCase):
    def test_registry(self) -> None:
        functions: decorators.Registry[tp.Callable[[], int]] = decorators.Registry()
        other: decorators.Registry[tp.Callable[[], int]] = decorators.Registry()

        @functions.register
        def dummy() -> int:
            return 12

        np.testing.assert_equal(dummy(), 12)
        np.testing.assert_array_equal(list(functions.keys()), ["dummy"])
        np.testing.assert_array_equal(list(other.keys()), [])
        functions.unregister("dummy")
        functions.unregister("other_dummy_that_does_not_exist")
        np.testing.assert_array_equal(list(functions.keys()), [])

    def test_info_registry(self) -> None:
        functions: decorators.Registry[tp.Callable[[], int]] = decorators.Registry()

        @functions.register_with_info(tag="info")
        def dummy_info() -> int:
            return 10

        np.testing.assert_equal(dummy_info(), 10)
        np.testing.assert_equal(functions.get_info("dummy_info"), {"tag": "info"})
        np.testing.assert_raises(ValueError, functions.get_info, "no_dummy")

    def test_registry_error(self) -> None:
        functions: decorators.Registry[tp.Any] = decorators.Registry()

        @functions.register
        def dummy() -> int:
            return 12

        np.testing.assert_raises(RuntimeError, functions.register, dummy)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""Groups of optimizers for use in benchmarks
"""
import typing as tp
import numpy as np
import nevergrad as ng
from nevergrad.common.decorators import Registry

from nevergrad.optimization import base as obase
from nevergrad.optimization.optimizerlib import ConfSplitOptimizer
from nevergrad.optimization.optimizerlib import registry as optimizerlib_registry
from nevergrad.optimization.optimizerlib import ParametrizedOnePlusOne
from nevergrad.optimization.optimizerlib import NonObjectOptimizer

Optim = tp.Union[obase.ConfiguredOptimizer, str]
registry: Registry[tp.Callable[[], tp.Iterable[Optim]]] = Registry()


def get_optimizers(*names: str, seed: tp.Optional[int] = None) -> tp.List[Optim]:
    """Returns an deterministically ordered list of optimizers, belonging to the
    provided groups. If a seed is provided, it is used to shuffle the list.

    Parameters
    ----------
    *names: str
        names of the groups to use. See nevergrad/benchmarks/optimizer_groups.txt
        (generated/updated when running pytest)
        for the list of groups and what they contain
    seed: optional int
        a seed to shuffle the list of optimizers
    """
    optims: tp.List[Optim] = []
    for name in names:
        for optim in registry[name]():
            if optim not in optims:  # avoid duplicates
                optims.append(optim)
    if seed is not None:
        np.random.RandomState(seed).shuffle(optims)  # type: ignore
    return optims


@registry.register
def large() -> tp.Sequence[Optim]:
    return [
        "CMA",
        "PSO",
        "DE",
        "MiniDE",
        "QrDE",
        "MiniQrDE",
        "LhsDE",
        "OnePlusOne",
        "SQP",
        "Cobyla",
        "Powell",
        "TwoPointsDE",
        "OnePointDE",
        "AlmostRotationInvariantDE",
        "RotationInvariantDE",
        "Portfolio",
        "CMandAS2",
        "CM",
        "MultiCMA",
        "TripleCMA",
        "MultiScaleCMA",
        "RSQP",
        "RCobyla",
        "RPowell",
        "MetaModel",
        "PolyCMA",
    ]


@registry.register
def emna_variants() -> tp.Sequence[Optim]:
    return [
        "IsoEMNA",
        "NaiveIsoEMNA",
        "AnisoEMNA",
        "NaiveAnisoEMNA",
        "CMA",
        "NaiveTBPSA",
        "NaiveIsoEMNATBPSA",
        "IsoEMNATBPSA",
        "NaiveAnisoEMNATBPSA",
        "AnisoEMNATBPSA",
    ]


@registry.register
def splitters() -> tp.Sequence[Optim]:
    optims: tp.List[Optim] = []
    for num_optims in [None, 3, 5, 9, 13]:
        for str_optim in ["CMA", "ECMA", "DE", "TwoPointsDE"]:
            optim = optimizerlib_registry[str_optim]
            name = "Split" + str_optim + ("Auto" if num_optims is None else str(num_optims))
            opt = ConfSplitOptimizer(multivariate_optimizer=optim, num_optims=num_optims).set_name(name)
            optims.append(opt)
    return optims


@registry.register
def progressive() -> tp.Sequence[Optim]:
    optims: tp.List[Optim] = []
    for mutation in ["discrete", "gaussian"]:
        for num_optims in [None, 13, 10000]:
            name = (
                "Prog"
                + ("Disc" if mutation == "discrete" else "")
                + ("Auto" if num_optims is None else ("Inf" if num_optims == 10000 else str(num_optims)))
            )
            mv = ParametrizedOnePlusOne(noise_handling="optimistic", mutation=mutation)
            opt = ConfSplitOptimizer(
                num_optims=num_optims, progressive=True, multivariate_optimizer=mv
            ).set_name(name)
            optims.append(opt)
    return optims


@registry.register
def anisotropic_progressive() -> tp.Sequence[Optim]:
    optims: tp.List[Optim] = []
    for num_optims in [None, 3, 5, 9, 13]:
        for str_optim in [
            "CMA",
            "ECMA",
            "DE",
            "TwoPointsDE",
            "PSO",
            "NoisyRL2",
            "NoisyRL3",
            "NoisyRL1",
            "MixDeterministicRL",
        ]:
            optim = optimizerlib_registry[str_optim]
            name = "Prog" + str_optim + ("Auto" if num_optims is None else str(num_optims))
            opt = ConfSplitOptimizer(
                multivariate_optimizer=optim, num_optims=num_optims, progressive=True
            ).set_name(name)
            optims.append(opt)
    return optims


@registry.register
def basics() -> tp.Sequence[Optim]:
    return ["NGOpt10", "CMandAS2", "CMA", "DE", "MetaModel"]


@registry.register
def baselines() -> tp.Sequence[Optim]:
    # This list should not change. This is the basics for comparison.
    # No algorithm with unstable other dependency.
    return ["OnePlusOne", "DiscreteOnePlusOne", "NoisyDiscreteOnePlusOne", "PSO", "DE", "TwoPointsDE"]


@registry.register
def parallel_basics() -> tp.Sequence[Optim]:
    return ["NGOpt10", "CMandAS2", "CMA", "DE", "MetaModel"]


@registry.register
def cma() -> tp.Sequence[Optim]:
    return ["DiagonalCMA", "CMA"]


@registry.register
def competence_map() -> tp.Sequence[Optim]:
    return ["NGO", "Shiwa"]


@registry.register
def competitive() -> tp.Sequence[Optim]:
    return get_optimizers("cma", "competence_map") + [
        #        "MetaNGOpt10",
        "NGOpt",
        "NGOptRW",
        "GeneticDE",
        "NaiveTBPSA",
        "PSO",
        "DE",
        "LhsDE",
        "RandomSearch",
        "OnePlusOne",
        "TwoPointsDE",
    ]


@registry.register
def all_bo() -> tp.Sequence[Optim]:
    return sorted(x for x in ng.optimizers.registry if "BO" in x)


@registry.register
def all_nlopts() -> tp.Sequence[Optim]:
    list_nlopt_options = [
        "LN_SBPLX",
        "LN_PRAXIS",
        "GN_DIRECT",
        "GN_DIRECT_L",
        "GN_CRS2_LM",
        "GN_AGS",
        "GN_ISRES",
        "GN_ESCH",
        "LN_COBYLA",
        "LN_BOBYQA",
        "LN_NEWUOA_BOUND",
        "LN_NELDERMEAD",
    ]
    return [NonObjectOptimizer(method="NLOPT_" + x).set_name("NLOPT_" + x) for x in list_nlopt_options]


@registry.register
def discrete() -> tp.Sequence[Optim]:
    return [name for name in optimizerlib_registry.keys() if "iscrete" in name and "oisy" not in name]


@registry.register
def noisy() -> tp.Sequence[Optim]:
    return ["OptimisticDiscreteOnePlusOne", "OptimisticNoisyOnePlusOne", "TBPSA", "SPSA", "NGOpt10"]


@registry.register
def structure() -> tp.Sequence[Optim]:
    return ["RecES", "RecMixES", "RecMutDE", "ParametrizationDE"]


@registry.register
def small_discrete() -> tp.Sequence[Optim]:
    return [
        "DiscreteOnePlusOne",
        "RLSOnePlusOne",
        "Shiwa",
        "CMA",
        "PSO",
        "TwoPointsDE",
        "DE",
        "OnePlusOne",
        "AdaptiveDiscreteOnePlusOne",
        "CMandAS2",
        "PortfolioDiscreteOnePlusOne",
        "DoubleFastGADiscreteOnePlusOne",
        "MultiDiscrete",
        "DiscreteBSOOnePlusOne",
        "AnisotropicAdaptiveDiscreteOnePlusOne",
        "DiscreteLenglerOnePlusOne",
    ]


@registry.register
def scipy() -> tp.Sequence[Optim]:
    return ["RSQP", "RCobyla", "RPowell", "SQPCMA", "SQP", "Cobyla", "Powell"]


@registry.register
def es() -> tp.Sequence[Optim]:
    es = [
        ng.families.EvolutionStrategy(
            recombination_ratio=recomb, only_offsprings=only, popsize=pop, offsprings=pop * 5
        )
        for only in [True, False]
        for recomb in [0.1, 0.5]
        for pop in [20, 40, 80]
    ]
    return es


@registry.register
def multimodal() -> tp.Sequence[Optim]:
    return ["NaiveTBPSA", "MultiCMA", "TripleCMA", "MultiScaleCMA", "PolyCMA", "QORandomSearch"]


# TODO(oteytaud): we should simplify the following.
@registry.register
def oneshot() -> tp.Sequence[Optim]:
    return sorted(
        x
        for x, y in optimizerlib_registry.items()
        if y.one_shot
        and "4" not in x
        and "7" not in x
        and "LHS" not in x
        and "alton" not in x
        and ("ando" not in x or "QO" in x)
    )  # QORandomSearch is the only valid variant of RandomSearch.


@registry.register
def structured_moo() -> tp.Sequence[Optim]:
    return [
        "CMA",
        "NGOpt10",
        "MetaNGOpt10",
        "DE",
        "PSO",
        "RecES",
        "RecMixES",
        "RecMutDE",
        "ParametrizationDE",
    ]


@registry.register
def spsa() -> tp.Sequence[Optim]:
    # return sorted(x for x, y in ng.optimizers.registry.items() if (any(e in x for e in "TBPSA SPSA".split()) and "iscr" not in x))
    return ["NaiveTBPSA", "SPSA", "TBPSA"]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from ..parametrization import mutation as mutations
from ..parametrization._datalayers import Int as Int
from . import constraints as constraints


__all__ = ["mutations", "constraints", "Int"]
